@@ -1,40 +1,44 @@
 use crate::check::inappropriate_handshake_message;
-use crate::conn::{CommonState, ConnectionRandoms, State};
-use crate::error::Error;
+#[cfg(feature = "quic")]
+use crate::common_state::Protocol;
+#[cfg(feature = "secret_extraction")]
+use crate::common_state::Side;
+use crate::common_state::{CommonState, State};
+use crate::conn::ConnectionRandoms;
+use crate::enums::{
+    AlertDescription, ContentType, HandshakeType, ProtocolVersion, SignatureScheme,
+};
+use crate::error::{Error, InvalidMessage, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::{HandshakeHash, HandshakeHashBuffer};
 use crate::kx;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::base::{Payload, PayloadU8};
 use crate::msgs::ccs::ChangeCipherSpecPayload;
-use crate::msgs::codec::Codec;
+use crate::msgs::enums::ExtensionType;
 use crate::msgs::enums::KeyUpdateRequest;
-use crate::msgs::enums::{AlertDescription, NamedGroup, ProtocolVersion};
-use crate::msgs::enums::{ContentType, ExtensionType, HandshakeType, SignatureScheme};
-use crate::msgs::handshake::ClientExtension;
-use crate::msgs::handshake::DigitallySignedStruct;
-use crate::msgs::handshake::EncryptedExtensions;
 use crate::msgs::handshake::NewSessionTicketPayloadTLS13;
 use crate::msgs::handshake::{CertificateEntry, CertificatePayloadTLS13};
+use crate::msgs::handshake::{ClientExtension, ServerExtension};
 use crate::msgs::handshake::{HandshakeMessagePayload, HandshakePayload};
 use crate::msgs::handshake::{HasServerExtensions, ServerHelloPayload};
 use crate::msgs::handshake::{PresharedKeyIdentity, PresharedKeyOffer};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
+#[cfg(feature = "secret_extraction")]
+use crate::suites::PartiallyExtractedSecrets;
 use crate::tls13::key_schedule::{
     KeyScheduleEarly, KeyScheduleHandshake, KeySchedulePreHandshake, KeyScheduleTraffic,
 };
 use crate::tls13::Tls13CipherSuite;
-use crate::verify;
-#[cfg(feature = "quic")]
-use crate::{conn::Protocol, msgs::base::PayloadU16, quic};
+use crate::verify::{self, DigitallySignedStruct};
 use crate::{sign, KeyLog};
 
 use super::client_conn::ClientConnectionData;
 use super::hs::ClientContext;
 use crate::client::common::ServerCertDetails;
 use crate::client::common::{ClientAuthDetails, ClientHelloDetails};
-use crate::client::{hs, ClientConfig, ServerName, StoresClientSessions};
+use crate::client::{hs, ClientConfig, ClientSessionStore, ServerName};
 
 use crate::ticketer::TimeBase;
 use ring::constant_time;
@@ -79,13 +83,13 @@ pub(super) fn handle_server_hello(
         .ok_or_else(|| {
             cx.common
                 .send_fatal_alert(AlertDescription::MissingExtension);
-            Error::PeerMisbehavedError("missing key share".to_string())
+            Error::PeerMisbehaved(PeerMisbehaved::MissingKeyShare)
         })?;
 
     if our_key_share.group() != their_key_share.group {
         return Err(cx
             .common
-            .illegal_param("wrong group for key share"));
+            .illegal_param(PeerMisbehaved::WrongGroupForKeyShare));
     }
 
     let key_schedule_pre_handshake = if let (Some(selected_psk), Some(early_key_schedule)) =
@@ -95,9 +99,9 @@ pub(super) fn handle_server_hello(
             let resuming_suite = match suite.can_resume_from(resuming.suite()) {
                 Some(resuming) => resuming,
                 None => {
-                    return Err(cx
-                        .common
-                        .illegal_param("server resuming incompatible suite"));
+                    return Err(cx.common.illegal_param(
+                        PeerMisbehaved::ResumptionOfferedWithIncompatibleCipherSuite,
+                    ));
                 }
             };
 
@@ -106,21 +110,19 @@ pub(super) fn handle_server_hello(
             if cx.data.early_data.is_enabled() && resuming_suite != suite {
                 return Err(cx
                     .common
-                    .illegal_param("server varied suite with early data"));
+                    .illegal_param(PeerMisbehaved::EarlyDataOfferedWithVariedCipherSuite));
             }
 
             if selected_psk != 0 {
                 return Err(cx
                     .common
-                    .illegal_param("server selected invalid psk"));
+                    .illegal_param(PeerMisbehaved::SelectedInvalidPsk));
             }
 
             debug!("Resuming using PSK");
             // The key schedule has been initialized and set in fill_in_psk_binder()
         } else {
-            return Err(Error::PeerMisbehavedError(
-                "server selected unoffered psk".to_string(),
-            ));
+            return Err(PeerMisbehaved::SelectedUnofferedPsk.into());
         }
         KeySchedulePreHandshake::from(early_key_schedule)
     } else {
@@ -129,7 +131,7 @@ pub(super) fn handle_server_hello(
         cx.data.early_data.rejected();
         cx.common.early_traffic = false;
         resuming_session.take();
-        KeySchedulePreHandshake::new(suite.hkdf_algorithm)
+        KeySchedulePreHandshake::new(suite)
     };
 
     let key_schedule = our_key_share.complete(&their_key_share.payload.0, |secret| {
@@ -137,37 +139,25 @@ pub(super) fn handle_server_hello(
     })?;
 
     // Remember what KX group the server liked for next time.
-    save_kx_hint(&config, &server_name, their_key_share.group);
+    config
+        .resumption
+        .store
+        .set_kx_hint(&server_name, their_key_share.group);
 
     // If we change keying when a subsequent handshake message is being joined,
     // the two halves will have different record layer protections.  Disallow this.
     cx.common.check_aligned_handshake()?;
 
     let hash_at_client_recvd_server_hello = transcript.get_current_hash();
-
-    let (key_schedule, client_key, server_key) = key_schedule.derive_handshake_secrets(
+    let key_schedule = key_schedule.derive_client_handshake_secrets(
+        cx.data.early_data.is_enabled(),
         hash_at_client_recvd_server_hello,
+        suite,
         &*config.key_log,
         &randoms.client,
+        cx.common,
     );
 
-    // Decrypt with the peer's key, encrypt with our own key
-    cx.common
-        .record_layer
-        .set_message_decrypter(suite.derive_decrypter(&server_key));
-
-    if !cx.data.early_data.is_enabled() {
-        // Set the client encryption key for handshakes if early data is not used
-        cx.common
-            .record_layer
-            .set_message_encrypter(suite.derive_encrypter(&client_key));
-    }
-
-    #[cfg(feature = "quic")]
-    if cx.common.is_quic() {
-        cx.common.quic.hs_secrets = Some(quic::Secrets::new(client_key, server_key, suite, true));
-    }
-
     emit_fake_ccs(&mut sent_tls13_fake_ccs, cx.common);
 
     Ok(Box::new(ExpectEncryptedExtensions {
@@ -189,9 +179,7 @@ fn validate_server_hello(
     for ext in &server_hello.extensions {
         if !ALLOWED_PLAINTEXT_EXTS.contains(&ext.get_type()) {
             common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-            return Err(Error::PeerMisbehavedError(
-                "server sent unexpected cleartext ext".to_string(),
-            ));
+            return Err(PeerMisbehaved::UnexpectedCleartextExtension.into());
         }
     }
 
@@ -202,13 +190,10 @@ pub(super) fn initial_key_share(
     config: &ClientConfig,
     server_name: &ServerName,
 ) -> Result<kx::KeyExchange, Error> {
-    let key = persist::ClientSessionKey::hint_for_server_name(server_name);
-    let key_buf = key.get_encoding();
-
-    let maybe_value = config.session_storage.get(&key_buf);
-
-    let group = maybe_value
-        .and_then(|enc| NamedGroup::read_bytes(&enc))
+    let group = config
+        .resumption
+        .store
+        .kx_hint(server_name)
         .and_then(|group| kx::KeyExchange::choose(group, &config.kx_groups))
         .unwrap_or_else(|| {
             config
@@ -220,14 +205,6 @@ pub(super) fn initial_key_share(
     kx::KeyExchange::start(group).ok_or(Error::FailedToGetRandomBytes)
 }
 
-fn save_kx_hint(config: &ClientConfig, server_name: &ServerName, group: NamedGroup) {
-    let key = persist::ClientSessionKey::hint_for_server_name(server_name);
-
-    config
-        .session_storage
-        .put(key.get_encoding(), group.get_encoding());
-}
-
 /// This implements the horrifying TLS1.3 hack where PSK binders have a
 /// data dependency on the message they are contained within.
 pub(super) fn fill_in_psk_binder(
@@ -236,8 +213,8 @@ pub(super) fn fill_in_psk_binder(
     hmp: &mut HandshakeMessagePayload,
 ) -> KeyScheduleEarly {
     // We need to know the hash function of the suite we're trying to resume into.
-    let hkdf_alg = resuming.suite().hkdf_algorithm;
-    let suite_hash = resuming.suite().hash_algorithm();
+    let suite = resuming.suite();
+    let suite_hash = suite.hash_algorithm();
 
     // The binder is calculated over the clienthello, but doesn't include itself or its
     // length, or the length of its container.
@@ -246,7 +223,7 @@ pub(super) fn fill_in_psk_binder(
 
     // Run a fake key_schedule to simulate what the server will do if it chooses
     // to resume.
-    let key_schedule = KeyScheduleEarly::new(hkdf_alg, resuming.secret());
+    let key_schedule = KeyScheduleEarly::new(suite, resuming.secret());
     let real_binder = key_schedule.resumption_psk_binder_key_and_sign_verify_data(&handshake_hash);
 
     if let HandshakePayload::ClientHello(ref mut ch) = hmp.payload {
@@ -259,7 +236,6 @@ pub(super) fn fill_in_psk_binder(
 pub(super) fn prepare_resumption(
     config: &ClientConfig,
     cx: &mut ClientContext<'_>,
-    ticket: Vec<u8>,
     resuming_session: &persist::Retrieved<&persist::Tls13ClientSessionValue>,
     exts: &mut Vec<ClientExtension>,
     doing_retry: bool,
@@ -289,7 +265,8 @@ pub(super) fn prepare_resumption(
         .output_len;
     let binder = vec![0u8; binder_len];
 
-    let psk_identity = PresharedKeyIdentity::new(ticket, obfuscated_ticket_age);
+    let psk_identity =
+        PresharedKeyIdentity::new(resuming_session.ticket().to_vec(), obfuscated_ticket_age);
     let psk_ext = PresharedKeyOffer::new(psk_identity, binder);
     exts.push(ClientExtension::PresharedKey(psk_ext));
 }
@@ -307,17 +284,12 @@ pub(super) fn derive_early_traffic_secret(
     emit_fake_ccs(sent_tls13_fake_ccs, cx.common);
 
     let client_hello_hash = transcript_buffer.get_hash_given(resuming_suite.hash_algorithm(), &[]);
-    let client_early_traffic_secret =
-        early_key_schedule.client_early_traffic_secret(&client_hello_hash, key_log, client_random);
-    // Set early data encryption key
-    cx.common
-        .record_layer
-        .set_message_encrypter(resuming_suite.derive_encrypter(&client_early_traffic_secret));
-
-    #[cfg(feature = "quic")]
-    if cx.common.is_quic() {
-        cx.common.quic.early_secret = Some(client_early_traffic_secret);
-    }
+    early_key_schedule.client_early_traffic_secret(
+        &client_hello_hash,
+        key_log,
+        client_random,
+        cx.common,
+    );
 
     // Now the client can send encrypted early data
     cx.common.early_traffic = true;
@@ -343,19 +315,16 @@ pub(super) fn emit_fake_ccs(sent_tls13_fake_ccs: &mut bool, common: &mut CommonS
 fn validate_encrypted_extensions(
     common: &mut CommonState,
     hello: &ClientHelloDetails,
-    exts: &EncryptedExtensions,
+    exts: &Vec<ServerExtension>,
 ) -> Result<(), Error> {
     if exts.has_duplicate_extension() {
         common.send_fatal_alert(AlertDescription::DecodeError);
-        return Err(Error::PeerMisbehavedError(
-            "server sent duplicate encrypted extensions".to_string(),
-        ));
+        return Err(PeerMisbehaved::DuplicateEncryptedExtensions.into());
     }
 
     if hello.server_sent_unsolicited_extensions(exts, &[]) {
         common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-        let msg = "server sent unsolicited encrypted extension".to_string();
-        return Err(Error::PeerMisbehavedError(msg));
+        return Err(PeerMisbehaved::UnsolicitedEncryptedExtension.into());
     }
 
     for ext in exts {
@@ -363,8 +332,7 @@ fn validate_encrypted_extensions(
             || DISALLOWED_TLS13_EXTS.contains(&ext.get_type())
         {
             common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-            let msg = "server sent inappropriate encrypted extension".to_string();
-            return Err(Error::PeerMisbehavedError(msg));
+            return Err(PeerMisbehaved::DisallowedEncryptedExtension.into());
         }
     }
 
@@ -404,7 +372,7 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
                     None => {
                         return Err(cx
                             .common
-                            .missing_extension("QUIC transport parameters not found"));
+                            .missing_extension(PeerMisbehaved::MissingQuicTransportParameters));
                     }
                 }
             }
@@ -423,12 +391,8 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
 
             if was_early_traffic && !cx.common.early_traffic {
                 // If no early traffic, set the encryption key for handshakes
-                cx.common
-                    .record_layer
-                    .set_message_encrypter(
-                        self.suite
-                            .derive_encrypter(self.key_schedule.client_key()),
-                    );
+                self.key_schedule
+                    .set_handshake_encrypter(cx.common);
             }
 
             cx.common.peer_certificates = Some(
@@ -454,8 +418,7 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
             }))
         } else {
             if exts.early_data_extension_offered() {
-                let msg = "server sent early data extension without resumption".to_string();
-                return Err(Error::PeerMisbehavedError(msg));
+                return Err(PeerMisbehaved::EarlyDataExtensionWithoutResumption.into());
             }
             Ok(Box::new(ExpectCertificateOrCertReq {
                 config: self.config,
@@ -483,10 +446,14 @@ struct ExpectCertificateOrCertReq {
 impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
     fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
         match m.payload {
-            MessagePayload::Handshake(HandshakeMessagePayload {
-                payload: HandshakePayload::CertificateTLS13(..),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateTLS13(..),
+                        ..
+                    },
                 ..
-            }) => Box::new(ExpectCertificate {
+            } => Box::new(ExpectCertificate {
                 config: self.config,
                 server_name: self.server_name,
                 randoms: self.randoms,
@@ -497,10 +464,14 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
                 client_auth: None,
             })
             .handle(cx, m),
-            MessagePayload::Handshake(HandshakeMessagePayload {
-                payload: HandshakePayload::CertificateRequestTLS13(..),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::CertificateRequestTLS13(..),
+                        ..
+                    },
                 ..
-            }) => Box::new(ExpectCertificateRequest {
+            } => Box::new(ExpectCertificateRequest {
                 config: self.config,
                 server_name: self.server_name,
                 randoms: self.randoms,
@@ -553,7 +524,7 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
             warn!("Server sent non-empty certreq context");
             cx.common
                 .send_fatal_alert(AlertDescription::DecodeError);
-            return Err(Error::CorruptMessagePayload(ContentType::Handshake));
+            return Err(InvalidMessage::InvalidCertRequest.into());
         }
 
         let tls13_sign_schemes = sign::supported_sign_tls13();
@@ -569,9 +540,7 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
         if compat_sigschemes.is_empty() {
             cx.common
                 .send_fatal_alert(AlertDescription::HandshakeFailure);
-            return Err(Error::PeerIncompatibleError(
-                "server sent bad certreq schemes".to_string(),
-            ));
+            return Err(PeerIncompatible::NoCertificateRequestSignatureSchemesInCommon.into());
         }
 
         let client_auth = ClientAuthDetails::resolve(
@@ -621,7 +590,7 @@ impl State<ClientConnectionData> for ExpectCertificate {
             warn!("certificate with non-empty context during handshake");
             cx.common
                 .send_fatal_alert(AlertDescription::DecodeError);
-            return Err(Error::CorruptMessagePayload(ContentType::Handshake));
+            return Err(InvalidMessage::InvalidCertRequest.into());
         }
 
         if cert_chain.any_entry_has_duplicate_extension()
@@ -630,26 +599,24 @@ impl State<ClientConnectionData> for ExpectCertificate {
             warn!("certificate chain contains unsolicited/unknown extension");
             cx.common
                 .send_fatal_alert(AlertDescription::UnsupportedExtension);
-            return Err(Error::PeerMisbehavedError(
-                "bad cert chain extensions".to_string(),
-            ));
+            return Err(PeerMisbehaved::BadCertChainExtensions.into());
         }
 
         let server_cert = ServerCertDetails::new(
             cert_chain.convert(),
             cert_chain.get_end_entity_ocsp(),
-            cert_chain.get_end_entity_scts(),
+            cert_chain
+                .get_end_entity_scts()
+                .map(|scts| scts.to_vec()),
         );
 
         if let Some(sct_list) = server_cert.scts.as_ref() {
             if hs::sct_list_is_invalid(sct_list) {
-                let error_msg = "server sent invalid SCT list".to_string();
-                return Err(Error::PeerMisbehavedError(error_msg));
+                return Err(PeerMisbehaved::InvalidSctList.into());
             }
 
             if !self.may_send_sct_list {
-                let error_msg = "server sent unsolicited SCT list".to_string();
-                return Err(Error::PeerMisbehavedError(error_msg));
+                return Err(PeerMisbehaved::UnsolicitedSctList.into());
             }
         }
 
@@ -706,7 +673,10 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
                 &self.server_cert.ocsp_response,
                 now,
             )
-            .map_err(|err| hs::send_cert_error_alert(cx.common, err))?;
+            .map_err(|err| {
+                cx.common
+                    .send_cert_verify_error_alert(err)
+            })?;
 
         // 2. Verify their signature on the handshake.
         let handshake_hash = self.transcript.get_current_hash();
@@ -718,7 +688,10 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
                 &self.server_cert.cert_chain[0],
                 cert_verify,
             )
-            .map_err(|err| hs::send_cert_error_alert(cx.common, err))?;
+            .map_err(|err| {
+                cx.common
+                    .send_cert_verify_error_alert(err)
+            })?;
 
         cx.common.peer_certificates = Some(self.server_cert.cert_chain);
         self.transcript.add_message(&m);
@@ -760,7 +733,7 @@ fn emit_certificate_tls13(
 
     let m = Message {
         version: ProtocolVersion::TLSv1_3,
-        payload: MessagePayload::Handshake(HandshakeMessagePayload {
+        payload: MessagePayload::handshake(HandshakeMessagePayload {
             typ: HandshakeType::Certificate,
             payload: HandshakePayload::CertificateTLS13(cert_payload),
         }),
@@ -782,7 +755,7 @@ fn emit_certverify_tls13(
 
     let m = Message {
         version: ProtocolVersion::TLSv1_3,
-        payload: MessagePayload::Handshake(HandshakeMessagePayload {
+        payload: MessagePayload::handshake(HandshakeMessagePayload {
             typ: HandshakeType::CertificateVerify,
             payload: HandshakePayload::CertificateVerify(dss),
         }),
@@ -802,7 +775,7 @@ fn emit_finished_tls13(
 
     let m = Message {
         version: ProtocolVersion::TLSv1_3,
-        payload: MessagePayload::Handshake(HandshakeMessagePayload {
+        payload: MessagePayload::handshake(HandshakeMessagePayload {
             typ: HandshakeType::Finished,
             payload: HandshakePayload::Finished(verify_data_payload),
         }),
@@ -819,7 +792,7 @@ fn emit_end_of_early_data_tls13(transcript: &mut HandshakeHash, common: &mut Com
 
     let m = Message {
         version: ProtocolVersion::TLSv1_3,
-        payload: MessagePayload::Handshake(HandshakeMessagePayload {
+        payload: MessagePayload::handshake(HandshakeMessagePayload {
             typ: HandshakeType::EndOfEarlyData,
             payload: HandshakePayload::EndOfEarlyData,
         }),
@@ -869,12 +842,8 @@ impl State<ClientConnectionData> for ExpectFinished {
             emit_end_of_early_data_tls13(&mut st.transcript, cx.common);
             cx.common.early_traffic = false;
             cx.data.early_data.finished();
-            cx.common
-                .record_layer
-                .set_message_encrypter(
-                    st.suite
-                        .derive_encrypter(st.key_schedule.client_key()),
-                );
+            st.key_schedule
+                .set_handshake_encrypter(cx.common);
         }
 
         /* Send our authentication/finished messages.  These are still encrypted
@@ -902,50 +871,43 @@ impl State<ClientConnectionData> for ExpectFinished {
             }
         }
 
-        let (key_schedule_finished, client_key, server_key) = st
+        let (key_schedule_pre_finished, verify_data) = st
             .key_schedule
-            .into_traffic_with_client_finished_pending(
+            .into_pre_finished_client_traffic(
                 hash_after_handshake,
+                st.transcript.get_current_hash(),
                 &*st.config.key_log,
                 &st.randoms.client,
             );
-        let handshake_hash = st.transcript.get_current_hash();
-        let (key_schedule_traffic, verify_data, _) =
-            key_schedule_finished.sign_client_finish(&handshake_hash);
+
         emit_finished_tls13(&mut st.transcript, verify_data, cx.common);
 
+        /* We're now sure this server supports TLS1.3.  But if we run out of TLS1.3 tickets
+         * when connecting to it again, we definitely don't want to attempt a TLS1.2 resumption. */
+        st.config
+            .resumption
+            .store
+            .remove_tls12_session(&st.server_name);
+
         /* Now move to our application traffic keys. */
         cx.common.check_aligned_handshake()?;
-
-        cx.common
-            .record_layer
-            .set_message_decrypter(st.suite.derive_decrypter(&server_key));
-
-        cx.common
-            .record_layer
-            .set_message_encrypter(st.suite.derive_encrypter(&client_key));
-
+        let key_schedule_traffic = key_schedule_pre_finished.into_traffic(cx.common);
         cx.common.start_traffic();
 
         let st = ExpectTraffic {
-            session_storage: Arc::clone(&st.config.session_storage),
+            session_storage: Arc::clone(&st.config.resumption.store),
             server_name: st.server_name,
             suite: st.suite,
             transcript: st.transcript,
             key_schedule: key_schedule_traffic,
-            want_write_key_update: false,
             _cert_verified: st.cert_verified,
             _sig_verified: st.sig_verified,
             _fin_verified: fin,
         };
 
         #[cfg(feature = "quic")]
-        {
-            if cx.common.protocol == Protocol::Quic {
-                cx.common.quic.traffic_secrets =
-                    Some(quic::Secrets::new(client_key, server_key, st.suite, true));
-                return Ok(Box::new(ExpectQuicTraffic(st)));
-            }
+        if cx.common.is_quic() {
+            return Ok(Box::new(ExpectQuicTraffic(st)));
         }
 
         Ok(Box::new(st))
@@ -956,12 +918,11 @@ impl State<ClientConnectionData> for ExpectFinished {
 // In this state we can be sent tickets, key updates,
 // and application data.
 struct ExpectTraffic {
-    session_storage: Arc<dyn StoresClientSessions>,
+    session_storage: Arc<dyn ClientSessionStore>,
     server_name: ServerName,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleTraffic,
-    want_write_key_update: bool,
     _cert_verified: verify::ServerCertVerified,
     _sig_verified: verify::HandshakeSignatureValid,
     _fin_verified: verify::FinishedMessageVerified,
@@ -977,9 +938,7 @@ impl ExpectTraffic {
         if nst.has_duplicate_extension() {
             cx.common
                 .send_fatal_alert(AlertDescription::IllegalParameter);
-            return Err(Error::PeerMisbehavedError(
-                "peer sent duplicate NewSessionTicket extensions".into(),
-            ));
+            return Err(PeerMisbehaved::DuplicateNewSessionTicketExtensions.into());
         }
 
         let handshake_hash = self.transcript.get_current_hash();
@@ -996,7 +955,8 @@ impl ExpectTraffic {
             }
         };
 
-        let value = persist::Tls13ClientSessionValue::new(
+        #[allow(unused_mut)]
+        let mut value = persist::Tls13ClientSessionValue::new(
             self.suite,
             nst.ticket.0.clone(),
             secret,
@@ -1012,77 +972,48 @@ impl ExpectTraffic {
         );
 
         #[cfg(feature = "quic")]
-        if let Some(sz) = nst.get_max_early_data_size() {
-            if cx.common.protocol == Protocol::Quic && sz != 0 && sz != 0xffff_ffff {
-                return Err(Error::PeerMisbehavedError(
-                    "invalid max_early_data_size".into(),
-                ));
+        if cx.common.is_quic() {
+            if let Some(sz) = nst.get_max_early_data_size() {
+                if sz != 0 && sz != 0xffff_ffff {
+                    return Err(PeerMisbehaved::InvalidMaxEarlyDataSize.into());
+                }
+            }
+
+            if let Some(ref quic_params) = &cx.common.quic.params {
+                value.set_quic_params(quic_params);
             }
         }
 
-        let key = persist::ClientSessionKey::session_for_server_name(&self.server_name);
-        #[allow(unused_mut)]
-        let mut ticket = value.get_encoding();
-
-        #[cfg(feature = "quic")]
-        if let (Protocol::Quic, Some(ref quic_params)) =
-            (cx.common.protocol, &cx.common.quic.params)
-        {
-            PayloadU16::encode_slice(quic_params, &mut ticket);
-        }
-
-        let worked = self
-            .session_storage
-            .put(key.get_encoding(), ticket);
-
-        if worked {
-            debug!("Ticket saved");
-        } else {
-            debug!("Ticket not saved");
-        }
+        self.session_storage
+            .insert_tls13_ticket(&self.server_name, value);
         Ok(())
     }
 
     fn handle_key_update(
         &mut self,
         common: &mut CommonState,
-        kur: &KeyUpdateRequest,
+        key_update_request: &KeyUpdateRequest,
     ) -> Result<(), Error> {
         #[cfg(feature = "quic")]
         {
             if let Protocol::Quic = common.protocol {
                 common.send_fatal_alert(AlertDescription::UnexpectedMessage);
-                let msg = "KeyUpdate received in QUIC connection".to_string();
-                warn!("{}", msg);
-                return Err(Error::PeerMisbehavedError(msg));
+                warn!("KeyUpdate received in QUIC connection");
+                return Err(PeerMisbehaved::KeyUpdateReceivedInQuicConnection.into());
             }
         }
 
         // Mustn't be interleaved with other handshake messages.
         common.check_aligned_handshake()?;
 
-        match kur {
-            KeyUpdateRequest::UpdateNotRequested => {}
-            KeyUpdateRequest::UpdateRequested => {
-                self.want_write_key_update = true;
-            }
-            _ => {
-                common.send_fatal_alert(AlertDescription::IllegalParameter);
-                return Err(Error::CorruptMessagePayload(ContentType::Handshake));
-            }
+        if common.should_update_key(key_update_request)? {
+            self.key_schedule
+                .update_encrypter_and_notify(common);
         }
 
         // Update our read-side keys.
-        let new_read_key = self
-            .key_schedule
-            .next_server_application_traffic_secret();
-        common
-            .record_layer
-            .set_message_decrypter(
-                self.suite
-                    .derive_decrypter(&new_read_key),
-            );
-
+        self.key_schedule
+            .update_decrypter(common);
         Ok(())
     }
 }
@@ -1093,14 +1024,22 @@ impl State<ClientConnectionData> for ExpectTraffic {
             MessagePayload::ApplicationData(payload) => cx
                 .common
                 .take_received_plaintext(payload),
-            MessagePayload::Handshake(HandshakeMessagePayload {
-                payload: HandshakePayload::NewSessionTicketTLS13(ref new_ticket),
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::NewSessionTicketTLS13(ref new_ticket),
+                        ..
+                    },
                 ..
-            }) => self.handle_new_ticket_tls13(cx, new_ticket)?,
-            MessagePayload::Handshake(HandshakeMessagePayload {
-                payload: HandshakePayload::KeyUpdate(ref key_update),
+            } => self.handle_new_ticket_tls13(cx, new_ticket)?,
+            MessagePayload::Handshake {
+                parsed:
+                    HandshakeMessagePayload {
+                        payload: HandshakePayload::KeyUpdate(ref key_update),
+                        ..
+                    },
                 ..
-            }) => self.handle_key_update(cx.common, key_update)?,
+            } => self.handle_key_update(cx.common, key_update)?,
             payload => {
                 return Err(inappropriate_handshake_message(
                     &payload,
@@ -1123,18 +1062,10 @@ impl State<ClientConnectionData> for ExpectTraffic {
             .export_keying_material(output, label, context)
     }
 
-    fn perhaps_write_key_update(&mut self, common: &mut CommonState) {
-        if self.want_write_key_update {
-            self.want_write_key_update = false;
-            common.send_msg_encrypt(Message::build_key_update_notify().into());
-
-            let write_key = self
-                .key_schedule
-                .next_client_application_traffic_secret();
-            common
-                .record_layer
-                .set_message_encrypter(self.suite.derive_encrypter(&write_key));
-        }
+    #[cfg(feature = "secret_extraction")]
+    fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
+        self.key_schedule
+            .extract_secrets(Side::Client)
     }
 }
 
