@@ -1,15 +1,31 @@
-use super::base::{Payload, PayloadU16, PayloadU24, PayloadU8};
-use super::codec::{put_u16, Codec, Reader};
-use super::enums::*;
-use super::handshake::*;
+use crate::enums::{CipherSuite, HandshakeType, ProtocolVersion, SignatureScheme};
 use crate::key::Certificate;
+use crate::msgs::base::{Payload, PayloadU16, PayloadU24, PayloadU8};
+use crate::msgs::codec::{put_u16, Codec, Reader};
+use crate::msgs::enums::{
+    ClientCertificateType, Compression, ECCurveType, ECPointFormat, ExtensionType,
+    KeyUpdateRequest, NamedGroup, PSKKeyExchangeMode, ServerNameType,
+};
+use crate::msgs::handshake::{
+    CertReqExtension, CertificateEntry, CertificateExtension, CertificatePayloadTLS13,
+    CertificateRequestPayload, CertificateRequestPayloadTLS13, CertificateStatus,
+    CertificateStatusRequest, ClientExtension, ClientHelloPayload, ClientSessionTicket,
+    ConvertProtocolNameList, ConvertServerNameList, DistinguishedName, ECDHEServerKeyExchange,
+    ECParameters, HandshakeMessagePayload, HandshakePayload, HasServerExtensions,
+    HelloRetryExtension, HelloRetryRequest, KeyShareEntry, NewSessionTicketExtension,
+    NewSessionTicketPayload, NewSessionTicketPayloadTLS13, PresharedKeyBinder,
+    PresharedKeyIdentity, PresharedKeyOffer, ProtocolName, Random, Sct, ServerECDHParams,
+    ServerExtension, ServerHelloPayload, ServerKeyExchangePayload, SessionId, UnknownExtension,
+};
+use crate::verify::DigitallySignedStruct;
+
 use webpki::DnsNameRef;
 
 #[test]
 fn rejects_short_random() {
     let bytes = [0x01; 31];
     let mut rd = Reader::init(&bytes);
-    assert_eq!(Random::read(&mut rd), None);
+    assert!(Random::read(&mut rd).is_err());
 }
 
 #[test]
@@ -22,24 +38,32 @@ fn reads_random() {
     assert!(!rd.any_left());
 }
 
+#[test]
+fn debug_random() {
+    assert_eq!(
+        "0101010101010101010101010101010101010101010101010101010101010101",
+        format!("{:?}", Random::from([1; 32]))
+    );
+}
+
 #[test]
 fn rejects_truncated_sessionid() {
     let bytes = [32; 32];
     let mut rd = Reader::init(&bytes);
-    assert_eq!(SessionID::read(&mut rd), None);
+    assert!(SessionId::read(&mut rd).is_err());
 }
 
 #[test]
 fn rejects_sessionid_with_bad_length() {
     let bytes = [33; 33];
     let mut rd = Reader::init(&bytes);
-    assert_eq!(SessionID::read(&mut rd), None);
+    assert!(SessionId::read(&mut rd).is_err());
 }
 
 #[test]
 fn sessionid_with_different_lengths_are_unequal() {
-    let a = SessionID::read(&mut Reader::init(&[1u8, 1])).unwrap();
-    let b = SessionID::read(&mut Reader::init(&[2u8, 1, 2])).unwrap();
+    let a = SessionId::read(&mut Reader::init(&[1u8, 1])).unwrap();
+    let b = SessionId::read(&mut Reader::init(&[2u8, 1, 2])).unwrap();
     assert_ne!(a, b);
 }
 
@@ -47,7 +71,7 @@ fn sessionid_with_different_lengths_are_unequal() {
 fn accepts_short_sessionid() {
     let bytes = [1; 2];
     let mut rd = Reader::init(&bytes);
-    let sess = SessionID::read(&mut rd).unwrap();
+    let sess = SessionId::read(&mut rd).unwrap();
     println!("{:?}", sess);
 
     assert!(!sess.is_empty());
@@ -59,7 +83,7 @@ fn accepts_short_sessionid() {
 fn accepts_empty_sessionid() {
     let bytes = [0; 1];
     let mut rd = Reader::init(&bytes);
-    let sess = SessionID::read(&mut rd).unwrap();
+    let sess = SessionId::read(&mut rd).unwrap();
     println!("{:?}", sess);
 
     assert!(sess.is_empty());
@@ -67,6 +91,20 @@ fn accepts_empty_sessionid() {
     assert!(!rd.any_left());
 }
 
+#[test]
+fn debug_sessionid() {
+    let bytes = [
+        32, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+        1, 1, 1,
+    ];
+    let mut rd = Reader::init(&bytes);
+    let sess = SessionId::read(&mut rd).unwrap();
+    assert_eq!(
+        "0101010101010101010101010101010101010101010101010101010101010101",
+        format!("{:?}", sess)
+    );
+}
+
 #[test]
 fn can_roundtrip_unknown_client_ext() {
     let bytes = [0x12u8, 0x34u8, 0, 3, 1, 2, 3];
@@ -82,42 +120,42 @@ fn can_roundtrip_unknown_client_ext() {
 fn refuses_client_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x0b, 0x00, 0x04, 0x02, 0xf8, 0x01, 0x02];
     let mut rd = Reader::init(&bytes);
-    assert!(ClientExtension::read(&mut rd).is_none());
+    assert!(ClientExtension::read(&mut rd).is_err());
 }
 
 #[test]
 fn refuses_server_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x0b, 0x00, 0x04, 0x02, 0xf8, 0x01, 0x02];
     let mut rd = Reader::init(&bytes);
-    assert!(ServerExtension::read(&mut rd).is_none());
+    assert!(ServerExtension::read(&mut rd).is_err());
 }
 
 #[test]
 fn refuses_certificate_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x12, 0x00, 0x03, 0x00, 0x00, 0x01];
     let mut rd = Reader::init(&bytes);
-    assert!(CertificateExtension::read(&mut rd).is_none());
+    assert!(CertificateExtension::read(&mut rd).is_err());
 }
 
 #[test]
 fn refuses_certificate_req_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x0d, 0x00, 0x05, 0x00, 0x02, 0x01, 0x02, 0xff];
     let mut rd = Reader::init(&bytes);
-    assert!(CertReqExtension::read(&mut rd).is_none());
+    assert!(CertReqExtension::read(&mut rd).is_err());
 }
 
 #[test]
 fn refuses_helloreq_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x2b, 0x00, 0x03, 0x00, 0x00, 0x01];
     let mut rd = Reader::init(&bytes);
-    assert!(HelloRetryExtension::read(&mut rd).is_none());
+    assert!(HelloRetryExtension::read(&mut rd).is_err());
 }
 
 #[test]
 fn refuses_newsessionticket_ext_with_unparsed_bytes() {
     let bytes = [0x00u8, 0x2a, 0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x01];
     let mut rd = Reader::init(&bytes);
-    assert!(NewSessionTicketExtension::read(&mut rd).is_none());
+    assert!(NewSessionTicketExtension::read(&mut rd).is_err());
 }
 
 #[test]
@@ -200,25 +238,25 @@ fn can_roundtrip_multiname_sni() {
 #[test]
 fn rejects_truncated_sni() {
     let bytes = [0, 0, 0, 1, 0];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 2, 0, 1];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 3, 0, 1, 0];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 4, 0, 2, 0, 0];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 5, 0, 3, 0, 0, 0];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 5, 0, 3, 0, 0, 1];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 
     let bytes = [0, 0, 0, 6, 0, 4, 0, 0, 2, 0x68];
-    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_none());
+    assert!(ClientExtension::read(&mut Reader::init(&bytes)).is_err());
 }
 
 #[test]
@@ -249,7 +287,7 @@ fn can_roundtrip_psk_offer() {
     assert_eq!(psko.identities[0].identity.0, vec![0x99]);
     assert_eq!(psko.identities[0].obfuscated_ticket_age, 0x11223344);
     assert_eq!(psko.binders.len(), 1);
-    assert_eq!(psko.binders[0].0, vec![1, 2, 3]);
+    assert_eq!(psko.binders[0].as_ref(), &[1, 2, 3]);
     assert_eq!(psko.get_encoding(), bytes.to_vec());
 }
 
@@ -320,54 +358,21 @@ fn can_roundtrip_single_proto() {
     }
 }
 
-#[test]
-fn decomposed_signature_scheme_has_correct_mappings() {
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::RSA, HashAlgorithm::SHA1),
-        SignatureScheme::RSA_PKCS1_SHA1
-    );
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::RSA, HashAlgorithm::SHA256),
-        SignatureScheme::RSA_PKCS1_SHA256
-    );
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::RSA, HashAlgorithm::SHA384),
-        SignatureScheme::RSA_PKCS1_SHA384
-    );
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::RSA, HashAlgorithm::SHA512),
-        SignatureScheme::RSA_PKCS1_SHA512
-    );
-
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::ECDSA, HashAlgorithm::SHA256),
-        SignatureScheme::ECDSA_NISTP256_SHA256
-    );
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::ECDSA, HashAlgorithm::SHA384),
-        SignatureScheme::ECDSA_NISTP384_SHA384
-    );
-    assert_eq!(
-        SignatureScheme::make(SignatureAlgorithm::ECDSA, HashAlgorithm::SHA512),
-        SignatureScheme::ECDSA_NISTP521_SHA512
-    );
-}
-
 fn get_sample_clienthellopayload() -> ClientHelloPayload {
     ClientHelloPayload {
         client_version: ProtocolVersion::TLSv1_2,
         random: Random::from([0; 32]),
-        session_id: SessionID::empty(),
+        session_id: SessionId::empty(),
         cipher_suites: vec![CipherSuite::TLS_NULL_WITH_NULL_NULL],
         compression_methods: vec![Compression::Null],
         extensions: vec![
-            ClientExtension::ECPointFormats(ECPointFormatList::supported()),
+            ClientExtension::ECPointFormats(ECPointFormat::SUPPORTED.to_vec()),
             ClientExtension::NamedGroups(vec![NamedGroup::X25519]),
             ClientExtension::SignatureAlgorithms(vec![SignatureScheme::ECDSA_NISTP256_SHA256]),
             ClientExtension::make_sni(DnsNameRef::try_from_ascii_str("hello").unwrap()),
             ClientExtension::SessionTicket(ClientSessionTicket::Request),
             ClientExtension::SessionTicket(ClientSessionTicket::Offer(Payload(vec![]))),
-            ClientExtension::Protocols(vec![PayloadU8(vec![0])]),
+            ClientExtension::Protocols(vec![ProtocolName::from(vec![0])]),
             ClientExtension::SupportedVersions(vec![ProtocolVersion::TLSv1_3]),
             ClientExtension::KeyShare(vec![KeyShareEntry::new(NamedGroup::X25519, &[1, 2, 3])]),
             ClientExtension::PresharedKeyModes(vec![PSKKeyExchangeMode::PSK_DHE_KE]),
@@ -377,8 +382,8 @@ fn get_sample_clienthellopayload() -> ClientHelloPayload {
                     PresharedKeyIdentity::new(vec![6, 7, 8], 7891011),
                 ],
                 binders: vec![
-                    PresharedKeyBinder::new(vec![1, 2, 3]),
-                    PresharedKeyBinder::new(vec![3, 4, 5]),
+                    PresharedKeyBinder::from(vec![1, 2, 3]),
+                    PresharedKeyBinder::from(vec![3, 4, 5]),
                 ],
             }),
             ClientExtension::Cookie(PayloadU16(vec![1, 2, 3])),
@@ -420,7 +425,7 @@ fn client_has_duplicate_extensions_works() {
 fn test_truncated_psk_offer() {
     let ext = ClientExtension::PresharedKey(PresharedKeyOffer {
         identities: vec![PresharedKeyIdentity::new(vec![3, 4, 5], 123456)],
-        binders: vec![PresharedKeyBinder::new(vec![1, 2, 3])],
+        binders: vec![PresharedKeyBinder::from(vec![1, 2, 3])],
     });
 
     let mut enc = ext.get_encoding();
@@ -431,7 +436,7 @@ fn test_truncated_psk_offer() {
         }
         put_u16(l as u16, &mut enc[4..]);
         let rc = ClientExtension::read_bytes(&enc);
-        assert!(rc.is_none());
+        assert!(rc.is_err());
     }
 }
 
@@ -446,7 +451,7 @@ fn test_truncated_client_hello_is_detected() {
         if l == 41 {
             continue; // where extensions are empty
         }
-        assert!(ClientHelloPayload::read_bytes(&enc[..l]).is_none());
+        assert!(ClientHelloPayload::read_bytes(&enc[..l]).is_err());
     }
 }
 
@@ -461,7 +466,7 @@ fn test_truncated_client_extension_is_detected() {
         // "outer" truncation, i.e., where the extension-level length is longer than
         // the input
         for l in 0..enc.len() {
-            assert!(ClientExtension::read_bytes(&enc[..l]).is_none());
+            assert!(ClientExtension::read_bytes(&enc[..l]).is_err());
         }
 
         // these extension types don't have any internal encoding that rustls validates:
@@ -477,7 +482,7 @@ fn test_truncated_client_extension_is_detected() {
         for l in 0..(enc.len() - 4) {
             put_u16(l as u16, &mut enc[2..]);
             println!("  encoding {:?} len {:?}", enc, l);
-            assert!(ClientExtension::read_bytes(&enc).is_none());
+            assert!(ClientExtension::read_bytes(&enc).is_err());
         }
     }
 }
@@ -580,7 +585,7 @@ fn test_truncated_helloretry_extension_is_detected() {
         // "outer" truncation, i.e., where the extension-level length is longer than
         // the input
         for l in 0..enc.len() {
-            assert!(HelloRetryExtension::read_bytes(&enc[..l]).is_none());
+            assert!(HelloRetryExtension::read_bytes(&enc[..l]).is_err());
         }
 
         // these extension types don't have any internal encoding that rustls validates:
@@ -593,7 +598,7 @@ fn test_truncated_helloretry_extension_is_detected() {
         for l in 0..(enc.len() - 4) {
             put_u16(l as u16, &mut enc[2..]);
             println!("  encoding {:?} len {:?}", enc, l);
-            assert!(HelloRetryExtension::read_bytes(&enc).is_none());
+            assert!(HelloRetryExtension::read_bytes(&enc).is_err());
         }
     }
 }
@@ -646,7 +651,7 @@ fn test_truncated_server_extension_is_detected() {
         // "outer" truncation, i.e., where the extension-level length is longer than
         // the input
         for l in 0..enc.len() {
-            assert!(ServerExtension::read_bytes(&enc[..l]).is_none());
+            assert!(ServerExtension::read_bytes(&enc[..l]).is_err());
         }
 
         // these extension types don't have any internal encoding that rustls validates:
@@ -662,7 +667,7 @@ fn test_truncated_server_extension_is_detected() {
         for l in 0..(enc.len() - 4) {
             put_u16(l as u16, &mut enc[2..]);
             println!("  encoding {:?} len {:?}", enc, l);
-            assert!(ServerExtension::read_bytes(&enc).is_none());
+            assert!(ServerExtension::read_bytes(&enc).is_err());
         }
     }
 }
@@ -750,20 +755,20 @@ fn get_sample_serverhellopayload() -> ServerHelloPayload {
     ServerHelloPayload {
         legacy_version: ProtocolVersion::TLSv1_2,
         random: Random::from([0; 32]),
-        session_id: SessionID::empty(),
+        session_id: SessionId::empty(),
         cipher_suite: CipherSuite::TLS_NULL_WITH_NULL_NULL,
         compression_method: Compression::Null,
         extensions: vec![
-            ServerExtension::ECPointFormats(ECPointFormatList::supported()),
+            ServerExtension::ECPointFormats(ECPointFormat::SUPPORTED.to_vec()),
             ServerExtension::ServerNameAck,
             ServerExtension::SessionTicketAck,
             ServerExtension::RenegotiationInfo(PayloadU8(vec![0])),
-            ServerExtension::Protocols(vec![PayloadU8(vec![0])]),
+            ServerExtension::Protocols(vec![ProtocolName::from(vec![0])]),
             ServerExtension::KeyShare(KeyShareEntry::new(NamedGroup::X25519, &[1, 2, 3])),
             ServerExtension::PresharedKey(3),
             ServerExtension::ExtendedMasterSecretAck,
             ServerExtension::CertificateStatusAck,
-            ServerExtension::SignedCertificateTimestamp(vec![PayloadU16(vec![0])]),
+            ServerExtension::SignedCertificateTimestamp(vec![Sct::from(vec![0])]),
             ServerExtension::SupportedVersions(ProtocolVersion::TLSv1_2),
             ServerExtension::TransportParameters(vec![1, 2, 3]),
             ServerExtension::Unknown(UnknownExtension {
@@ -787,7 +792,7 @@ fn can_clone_all_serverextensions() {
 fn get_sample_helloretryrequest() -> HelloRetryRequest {
     HelloRetryRequest {
         legacy_version: ProtocolVersion::TLSv1_2,
-        session_id: SessionID::empty(),
+        session_id: SessionId::empty(),
         cipher_suite: CipherSuite::TLS_NULL_WITH_NULL_NULL,
         extensions: vec![
             HelloRetryExtension::KeyShare(NamedGroup::X25519),
@@ -810,7 +815,7 @@ fn get_sample_certificatepayloadtls13() -> CertificatePayloadTLS13 {
                 CertificateExtension::CertificateStatus(CertificateStatus {
                     ocsp_response: PayloadU24(vec![1, 2, 3]),
                 }),
-                CertificateExtension::SignedCertificateTimestamp(vec![PayloadU16(vec![0])]),
+                CertificateExtension::SignedCertificateTimestamp(vec![Sct::from(vec![0])]),
                 CertificateExtension::Unknown(UnknownExtension {
                     typ: ExtensionType::Unknown(12345),
                     payload: Payload(vec![1, 2, 3]),
@@ -829,10 +834,7 @@ fn get_sample_serverkeyexchangepayload_ecdhe() -> ServerKeyExchangePayload {
             },
             public: PayloadU8(vec![1, 2, 3]),
         },
-        dss: DigitallySignedStruct {
-            scheme: SignatureScheme::RSA_PSS_SHA256,
-            sig: PayloadU16(vec![1, 2, 3]),
-        },
+        dss: DigitallySignedStruct::new(SignatureScheme::RSA_PSS_SHA256, vec![1, 2, 3]),
     })
 }
 
@@ -844,7 +846,7 @@ fn get_sample_certificaterequestpayload() -> CertificateRequestPayload {
     CertificateRequestPayload {
         certtypes: vec![ClientCertificateType::RSASign],
         sigschemes: vec![SignatureScheme::ECDSA_NISTP256_SHA256],
-        canames: vec![PayloadU16(vec![1, 2, 3])],
+        canames: vec![DistinguishedName::from(vec![1, 2, 3])],
     }
 }
 
@@ -853,7 +855,7 @@ fn get_sample_certificaterequestpayloadtls13() -> CertificateRequestPayloadTLS13
         context: PayloadU8(vec![1, 2, 3]),
         extensions: vec![
             CertReqExtension::SignatureAlgorithms(vec![SignatureScheme::ECDSA_NISTP256_SHA256]),
-            CertReqExtension::AuthorityNames(vec![PayloadU16(vec![1, 2, 3])]),
+            CertReqExtension::AuthorityNames(vec![DistinguishedName::from(vec![1, 2, 3])]),
             CertReqExtension::Unknown(UnknownExtension {
                 typ: ExtensionType::Unknown(12345),
                 payload: Payload(vec![1, 2, 3]),
@@ -882,7 +884,7 @@ fn get_sample_newsessionticketpayloadtls13() -> NewSessionTicketPayloadTLS13 {
     }
 }
 
-fn get_sample_encryptedextensions() -> EncryptedExtensions {
+fn get_sample_encryptedextensions() -> Vec<ServerExtension> {
     get_sample_serverhellopayload().extensions
 }
 
@@ -992,7 +994,7 @@ fn can_detect_truncation_of_all_tls12_handshake_payloads() {
 
         // outer truncation
         for l in 0..enc.len() {
-            assert!(HandshakeMessagePayload::read_bytes(&enc[..l]).is_none())
+            assert!(HandshakeMessagePayload::read_bytes(&enc[..l]).is_err())
         }
 
         // inner truncation
@@ -1014,8 +1016,8 @@ fn can_detect_truncation_of_all_tls12_handshake_payloads() {
                 &mut Reader::init(&enc),
                 ProtocolVersion::TLSv1_2
             )
-            .is_none());
-            assert!(HandshakeMessagePayload::read_bytes(&enc).is_none());
+            .is_err());
+            assert!(HandshakeMessagePayload::read_bytes(&enc).is_err());
         }
     }
 }
@@ -1139,7 +1141,7 @@ fn can_detect_truncation_of_all_tls13_handshake_payloads() {
 
         // outer truncation
         for l in 0..enc.len() {
-            assert!(HandshakeMessagePayload::read_bytes(&enc[..l]).is_none())
+            assert!(HandshakeMessagePayload::read_bytes(&enc[..l]).is_err())
         }
 
         // inner truncation
@@ -1161,7 +1163,7 @@ fn can_detect_truncation_of_all_tls13_handshake_payloads() {
                 &mut Reader::init(&enc),
                 ProtocolVersion::TLSv1_3
             )
-            .is_none());
+            .is_err());
         }
     }
 }
@@ -1174,7 +1176,7 @@ fn cannot_read_messagehash_from_network() {
     };
     println!("mh {:?}", mh);
     let enc = mh.get_encoding();
-    assert!(HandshakeMessagePayload::read_bytes(&enc).is_none());
+    assert!(HandshakeMessagePayload::read_bytes(&enc).is_err());
 }
 
 #[test]
@@ -1200,5 +1202,13 @@ fn cannot_decode_huge_certificate() {
     buf[4] = 0x01;
     buf[5] = 0x00;
     buf[6] = 0x01;
-    assert!(HandshakeMessagePayload::read_bytes(&buf).is_none());
+    assert!(HandshakeMessagePayload::read_bytes(&buf).is_err());
+}
+
+#[test]
+fn can_decode_server_hello_from_api_devicecheck_apple_com() {
+    let data = include_bytes!("hello-api.devicecheck.apple.com.bin");
+    let mut r = Reader::init(data);
+    let hm = HandshakeMessagePayload::read(&mut r).unwrap();
+    println!("msg: {:?}", hm);
 }
