@@ -1,16 +1,428 @@
 /// This module contains optional APIs for implementing QUIC TLS.
 use crate::cipher::{Iv, IvLen};
-pub use crate::client::ClientQuicExt;
-use crate::conn::CommonState;
+use crate::client::{ClientConfig, ClientConnectionData, ServerName};
+use crate::common_state::{CommonState, Protocol, Side};
+use crate::conn::{ConnectionCore, SideData};
+use crate::enums::{AlertDescription, ProtocolVersion};
 use crate::error::Error;
-use crate::msgs::enums::AlertDescription;
-pub use crate::server::ServerQuicExt;
+use crate::msgs::handshake::{ClientExtension, ServerExtension};
+use crate::server::{ServerConfig, ServerConnectionData};
 use crate::suites::BulkAlgorithm;
 use crate::tls13::key_schedule::hkdf_expand;
 use crate::tls13::{Tls13CipherSuite, TLS13_AES_128_GCM_SHA256_INTERNAL};
 
 use ring::{aead, hkdf};
 
+use std::collections::VecDeque;
+use std::fmt::{self, Debug};
+use std::ops::{Deref, DerefMut};
+use std::sync::Arc;
+
+/// A QUIC client or server connection.
+#[derive(Debug)]
+pub enum Connection {
+    /// A client connection
+    Client(ClientConnection),
+    /// A server connection
+    Server(ServerConnection),
+}
+
+impl Connection {
+    /// Return the TLS-encoded transport parameters for the session's peer.
+    ///
+    /// See [`ConnectionCommon::quic_transport_parameters()`] for more details.
+    pub fn quic_transport_parameters(&self) -> Option<&[u8]> {
+        match self {
+            Self::Client(conn) => conn.quic_transport_parameters(),
+            Self::Server(conn) => conn.quic_transport_parameters(),
+        }
+    }
+
+    /// Compute the keys for encrypting/decrypting 0-RTT packets, if available
+    pub fn zero_rtt_keys(&self) -> Option<DirectionalKeys> {
+        match self {
+            Self::Client(conn) => conn.zero_rtt_keys(),
+            Self::Server(conn) => conn.zero_rtt_keys(),
+        }
+    }
+
+    /// Consume unencrypted TLS handshake data.
+    ///
+    /// Handshake data obtained from separate encryption levels should be supplied in separate calls.
+    pub fn read_hs(&mut self, plaintext: &[u8]) -> Result<(), Error> {
+        match self {
+            Self::Client(conn) => conn.read_hs(plaintext),
+            Self::Server(conn) => conn.read_hs(plaintext),
+        }
+    }
+
+    /// Emit unencrypted TLS handshake data.
+    ///
+    /// When this returns `Some(_)`, the new keys must be used for future handshake data.
+    pub fn write_hs(&mut self, buf: &mut Vec<u8>) -> Option<KeyChange> {
+        match self {
+            Self::Client(conn) => conn.write_hs(buf),
+            Self::Server(conn) => conn.write_hs(buf),
+        }
+    }
+
+    /// Emit the TLS description code of a fatal alert, if one has arisen.
+    ///
+    /// Check after `read_hs` returns `Err(_)`.
+    pub fn alert(&self) -> Option<AlertDescription> {
+        match self {
+            Self::Client(conn) => conn.alert(),
+            Self::Server(conn) => conn.alert(),
+        }
+    }
+
+    /// Derives key material from the agreed connection secrets.
+    ///
+    /// This function fills in `output` with `output.len()` bytes of key
+    /// material derived from the master session secret using `label`
+    /// and `context` for diversification. Ownership of the buffer is taken
+    /// by the function and returned via the Ok result to ensure no key
+    /// material leaks if the function fails.
+    ///
+    /// See RFC5705 for more details on what this does and is for.
+    ///
+    /// For TLS1.3 connections, this function does not use the
+    /// "early" exporter at any point.
+    ///
+    /// This function fails if called prior to the handshake completing;
+    /// check with [`CommonState::is_handshaking`] first.
+    #[inline]
+    pub fn export_keying_material<T: AsMut<[u8]>>(
+        &self,
+        output: T,
+        label: &[u8],
+        context: Option<&[u8]>,
+    ) -> Result<T, Error> {
+        match self {
+            Self::Client(conn) => conn
+                .core
+                .export_keying_material(output, label, context),
+            Self::Server(conn) => conn
+                .core
+                .export_keying_material(output, label, context),
+        }
+    }
+}
+
+impl Deref for Connection {
+    type Target = CommonState;
+
+    fn deref(&self) -> &Self::Target {
+        match self {
+            Self::Client(conn) => &conn.core.common_state,
+            Self::Server(conn) => &conn.core.common_state,
+        }
+    }
+}
+
+impl DerefMut for Connection {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        match self {
+            Self::Client(conn) => &mut conn.core.common_state,
+            Self::Server(conn) => &mut conn.core.common_state,
+        }
+    }
+}
+
+/// A QUIC client connection.
+pub struct ClientConnection {
+    inner: ConnectionCommon<ClientConnectionData>,
+}
+
+impl ClientConnection {
+    /// Make a new QUIC ClientConnection. This differs from `ClientConnection::new()`
+    /// in that it takes an extra argument, `params`, which contains the
+    /// TLS-encoded transport parameters to send.
+    pub fn new(
+        config: Arc<ClientConfig>,
+        quic_version: Version,
+        name: ServerName,
+        params: Vec<u8>,
+    ) -> Result<Self, Error> {
+        if !config.supports_version(ProtocolVersion::TLSv1_3) {
+            return Err(Error::General(
+                "TLS 1.3 support is required for QUIC".into(),
+            ));
+        }
+
+        let ext = match quic_version {
+            Version::V1Draft => ClientExtension::TransportParametersDraft(params),
+            Version::V1 => ClientExtension::TransportParameters(params),
+        };
+
+        Ok(Self {
+            inner: ConnectionCore::for_client(config, name, vec![ext], Protocol::Quic)?.into(),
+        })
+    }
+
+    /// Returns True if the server signalled it will process early data.
+    ///
+    /// If you sent early data and this returns false at the end of the
+    /// handshake then the server will not process the data.  This
+    /// is not an error, but you may wish to resend the data.
+    pub fn is_early_data_accepted(&self) -> bool {
+        self.inner.core.is_early_data_accepted()
+    }
+}
+
+impl Deref for ClientConnection {
+    type Target = ConnectionCommon<ClientConnectionData>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.inner
+    }
+}
+
+impl DerefMut for ClientConnection {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.inner
+    }
+}
+
+impl Debug for ClientConnection {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_struct("quic::ClientConnection")
+            .finish()
+    }
+}
+
+impl From<ClientConnection> for Connection {
+    fn from(c: ClientConnection) -> Self {
+        Self::Client(c)
+    }
+}
+
+/// A QUIC server connection.
+pub struct ServerConnection {
+    inner: ConnectionCommon<ServerConnectionData>,
+}
+
+impl ServerConnection {
+    /// Make a new QUIC ServerConnection. This differs from `ServerConnection::new()`
+    /// in that it takes an extra argument, `params`, which contains the
+    /// TLS-encoded transport parameters to send.
+    pub fn new(
+        config: Arc<ServerConfig>,
+        quic_version: Version,
+        params: Vec<u8>,
+    ) -> Result<Self, Error> {
+        if !config.supports_version(ProtocolVersion::TLSv1_3) {
+            return Err(Error::General(
+                "TLS 1.3 support is required for QUIC".into(),
+            ));
+        }
+
+        if config.max_early_data_size != 0 && config.max_early_data_size != 0xffff_ffff {
+            return Err(Error::General(
+                "QUIC sessions must set a max early data of 0 or 2^32-1".into(),
+            ));
+        }
+
+        let ext = match quic_version {
+            Version::V1Draft => ServerExtension::TransportParametersDraft(params),
+            Version::V1 => ServerExtension::TransportParameters(params),
+        };
+
+        let mut core = ConnectionCore::for_server(config, vec![ext])?;
+        core.common_state.protocol = Protocol::Quic;
+        Ok(Self { inner: core.into() })
+    }
+
+    /// Explicitly discard early data, notifying the client
+    ///
+    /// Useful if invariants encoded in `received_resumption_data()` cannot be respected.
+    ///
+    /// Must be called while `is_handshaking` is true.
+    pub fn reject_early_data(&mut self) {
+        self.inner.core.reject_early_data()
+    }
+
+    /// Retrieves the server name, if any, used to select the certificate and
+    /// private key.
+    ///
+    /// This returns `None` until some time after the client's server name indication
+    /// (SNI) extension value is processed during the handshake. It will never be
+    /// `None` when the connection is ready to send or process application data,
+    /// unless the client does not support SNI.
+    ///
+    /// This is useful for application protocols that need to enforce that the
+    /// server name matches an application layer protocol hostname. For
+    /// example, HTTP/1.1 servers commonly expect the `Host:` header field of
+    /// every request on a connection to match the hostname in the SNI extension
+    /// when the client provides the SNI extension.
+    ///
+    /// The server name is also used to match sessions during session resumption.
+    pub fn server_name(&self) -> Option<&str> {
+        self.inner.core.get_sni_str()
+    }
+}
+
+impl Deref for ServerConnection {
+    type Target = ConnectionCommon<ServerConnectionData>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.inner
+    }
+}
+
+impl DerefMut for ServerConnection {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.inner
+    }
+}
+
+impl Debug for ServerConnection {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_struct("quic::ServerConnection")
+            .finish()
+    }
+}
+
+impl From<ServerConnection> for Connection {
+    fn from(c: ServerConnection) -> Self {
+        Self::Server(c)
+    }
+}
+
+/// A shared interface for QUIC connections.
+pub struct ConnectionCommon<Data> {
+    core: ConnectionCore<Data>,
+}
+
+impl<Data: SideData> ConnectionCommon<Data> {
+    /// Return the TLS-encoded transport parameters for the session's peer.
+    ///
+    /// While the transport parameters are technically available prior to the
+    /// completion of the handshake, they cannot be fully trusted until the
+    /// handshake completes, and reliance on them should be minimized.
+    /// However, any tampering with the parameters will cause the handshake
+    /// to fail.
+    pub fn quic_transport_parameters(&self) -> Option<&[u8]> {
+        self.core
+            .common_state
+            .quic
+            .params
+            .as_ref()
+            .map(|v| v.as_ref())
+    }
+
+    /// Compute the keys for encrypting/decrypting 0-RTT packets, if available
+    pub fn zero_rtt_keys(&self) -> Option<DirectionalKeys> {
+        Some(DirectionalKeys::new(
+            self.core
+                .common_state
+                .suite
+                .and_then(|suite| suite.tls13())?,
+            self.core
+                .common_state
+                .quic
+                .early_secret
+                .as_ref()?,
+        ))
+    }
+
+    /// Consume unencrypted TLS handshake data.
+    ///
+    /// Handshake data obtained from separate encryption levels should be supplied in separate calls.
+    pub fn read_hs(&mut self, plaintext: &[u8]) -> Result<(), Error> {
+        self.core
+            .message_deframer
+            .push(ProtocolVersion::TLSv1_3, plaintext)?;
+        self.core.process_new_packets()?;
+        Ok(())
+    }
+
+    /// Emit unencrypted TLS handshake data.
+    ///
+    /// When this returns `Some(_)`, the new keys must be used for future handshake data.
+    pub fn write_hs(&mut self, buf: &mut Vec<u8>) -> Option<KeyChange> {
+        self.core
+            .common_state
+            .quic
+            .write_hs(buf)
+    }
+
+    /// Emit the TLS description code of a fatal alert, if one has arisen.
+    ///
+    /// Check after `read_hs` returns `Err(_)`.
+    pub fn alert(&self) -> Option<AlertDescription> {
+        self.core.common_state.quic.alert
+    }
+}
+
+impl<Data> Deref for ConnectionCommon<Data> {
+    type Target = CommonState;
+
+    fn deref(&self) -> &Self::Target {
+        &self.core.common_state
+    }
+}
+
+impl<Data> DerefMut for ConnectionCommon<Data> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.core.common_state
+    }
+}
+
+impl<Data> From<ConnectionCore<Data>> for ConnectionCommon<Data> {
+    fn from(core: ConnectionCore<Data>) -> Self {
+        Self { core }
+    }
+}
+
+#[cfg(feature = "quic")]
+#[derive(Default)]
+pub(crate) struct Quic {
+    /// QUIC transport parameters received from the peer during the handshake
+    pub(crate) params: Option<Vec<u8>>,
+    pub(crate) alert: Option<AlertDescription>,
+    pub(crate) hs_queue: VecDeque<(bool, Vec<u8>)>,
+    pub(crate) early_secret: Option<ring::hkdf::Prk>,
+    pub(crate) hs_secrets: Option<Secrets>,
+    pub(crate) traffic_secrets: Option<Secrets>,
+    /// Whether keys derived from traffic_secrets have been passed to the QUIC implementation
+    pub(crate) returned_traffic_keys: bool,
+}
+
+impl Quic {
+    pub(crate) fn write_hs(&mut self, buf: &mut Vec<u8>) -> Option<KeyChange> {
+        while let Some((_, msg)) = self.hs_queue.pop_front() {
+            buf.extend_from_slice(&msg);
+            if let Some(&(true, _)) = self.hs_queue.front() {
+                if self.hs_secrets.is_some() {
+                    // Allow the caller to switch keys before proceeding.
+                    break;
+                }
+            }
+        }
+
+        if let Some(secrets) = self.hs_secrets.take() {
+            return Some(KeyChange::Handshake {
+                keys: Keys::new(&secrets),
+            });
+        }
+
+        if let Some(mut secrets) = self.traffic_secrets.take() {
+            if !self.returned_traffic_keys {
+                self.returned_traffic_keys = true;
+                let keys = Keys::new(&secrets);
+                secrets.update();
+                return Some(KeyChange::OneRtt {
+                    keys,
+                    next: secrets,
+                });
+            }
+        }
+
+        None
+    }
+}
+
 /// Secrets used to encrypt/decrypt traffic
 #[derive(Clone, Debug)]
 pub struct Secrets {
@@ -20,7 +432,7 @@ pub struct Secrets {
     server: hkdf::Prk,
     /// Cipher suite used with these secrets
     suite: &'static Tls13CipherSuite,
-    is_client: bool,
+    side: Side,
 }
 
 impl Secrets {
@@ -28,13 +440,13 @@ impl Secrets {
         client: hkdf::Prk,
         server: hkdf::Prk,
         suite: &'static Tls13CipherSuite,
-        is_client: bool,
+        side: Side,
     ) -> Self {
         Self {
             client,
             server,
             suite,
-            is_client,
+            side,
         }
     }
 
@@ -52,44 +464,13 @@ impl Secrets {
     }
 
     fn local_remote(&self) -> (&hkdf::Prk, &hkdf::Prk) {
-        if self.is_client {
-            (&self.client, &self.server)
-        } else {
-            (&self.server, &self.client)
+        match self.side {
+            Side::Client => (&self.client, &self.server),
+            Side::Server => (&self.server, &self.client),
         }
     }
 }
 
-/// Generic methods for QUIC sessions
-pub trait QuicExt {
-    /// Return the TLS-encoded transport parameters for the session's peer.
-    ///
-    /// While the transport parameters are technically available prior to the
-    /// completion of the handshake, they cannot be fully trusted until the
-    /// handshake completes, and reliance on them should be minimized.
-    /// However, any tampering with the parameters will cause the handshake
-    /// to fail.
-    fn quic_transport_parameters(&self) -> Option<&[u8]>;
-
-    /// Compute the keys for encrypting/decrypting 0-RTT packets, if available
-    fn zero_rtt_keys(&self) -> Option<DirectionalKeys>;
-
-    /// Consume unencrypted TLS handshake data.
-    ///
-    /// Handshake data obtained from separate encryption levels should be supplied in separate calls.
-    fn read_hs(&mut self, plaintext: &[u8]) -> Result<(), Error>;
-
-    /// Emit unencrypted TLS handshake data.
-    ///
-    /// When this returns `Some(_)`, the new keys must be used for future handshake data.
-    fn write_hs(&mut self, buf: &mut Vec<u8>) -> Option<KeyChange>;
-
-    /// Emit the TLS description code of a fatal alert, if one has arisen.
-    ///
-    /// Check after `read_hs` returns `Err(_)`.
-    fn alert(&self) -> Option<AlertDescription>;
-}
-
 /// Keys used to communicate in a single direction
 pub struct DirectionalKeys {
     /// Encrypts or decrypts a packet's headers
@@ -217,7 +598,7 @@ impl HeaderProtectionKey {
 
         let first_plain = match masked {
             // When unmasking, use the packet length bits after unmasking
-            true => (*first ^ (first_mask & bits)),
+            true => *first ^ (first_mask & bits),
             // When masking, use the packet length bits before masking
             false => *first,
         };
@@ -373,7 +754,7 @@ pub struct Keys {
 
 impl Keys {
     /// Construct keys for use with initial packets
-    pub fn initial(version: Version, client_dst_connection_id: &[u8], is_client: bool) -> Self {
+    pub fn initial(version: Version, client_dst_connection_id: &[u8], side: Side) -> Self {
         const CLIENT_LABEL: &[u8] = b"client in";
         const SERVER_LABEL: &[u8] = b"server in";
         let salt = version.initial_salt();
@@ -383,7 +764,7 @@ impl Keys {
             client: hkdf_expand(&hs_secret, hkdf::HKDF_SHA256, CLIENT_LABEL, &[]),
             server: hkdf_expand(&hs_secret, hkdf::HKDF_SHA256, SERVER_LABEL, &[]),
             suite: TLS13_AES_128_GCM_SHA256_INTERNAL,
-            is_client,
+            side,
         };
         Self::new(&secrets)
     }
@@ -397,47 +778,15 @@ impl Keys {
     }
 }
 
-pub(crate) fn write_hs(this: &mut CommonState, buf: &mut Vec<u8>) -> Option<KeyChange> {
-    while let Some((_, msg)) = this.quic.hs_queue.pop_front() {
-        buf.extend_from_slice(&msg);
-        if let Some(&(true, _)) = this.quic.hs_queue.front() {
-            if this.quic.hs_secrets.is_some() {
-                // Allow the caller to switch keys before proceeding.
-                break;
-            }
-        }
-    }
-
-    if let Some(secrets) = this.quic.hs_secrets.take() {
-        return Some(KeyChange::Handshake {
-            keys: Keys::new(&secrets),
-        });
-    }
-
-    if let Some(mut secrets) = this.quic.traffic_secrets.take() {
-        if !this.quic.returned_traffic_keys {
-            this.quic.returned_traffic_keys = true;
-            let keys = Keys::new(&secrets);
-            secrets.update();
-            return Some(KeyChange::OneRtt {
-                keys,
-                next: secrets,
-            });
-        }
-    }
-
-    None
-}
-
 /// Key material for use in QUIC packet spaces
 ///
 /// QUIC uses 4 different sets of keys (and progressive key updates for long-running connections):
 ///
 /// * Initial: these can be created from [`Keys::initial()`]
-/// * 0-RTT keys: can be retrieved from [`QuicExt::zero_rtt_keys()`]
-/// * Handshake: these are returned from [`QuicExt::write_hs()`] after `ClientHello` and
+/// * 0-RTT keys: can be retrieved from [`ConnectionCommon::zero_rtt_keys()`]
+/// * Handshake: these are returned from [`ConnectionCommon::write_hs()`] after `ClientHello` and
 ///   `ServerHello` messages have been exchanged
-/// * 1-RTT keys: these are returned from [`QuicExt::write_hs()`] after the handshake is done
+/// * 1-RTT keys: these are returned from [`ConnectionCommon::write_hs()`] after the handshake is done
 ///
 /// Once the 1-RTT keys have been exchanged, either side may initiate a key update. Progressive
 /// update keys can be obtained from the [`Secrets`] returned in [`KeyChange::OneRtt`]. Note that
@@ -472,7 +821,7 @@ fn nonce_for(packet_number: u64, iv: &Iv) -> ring::aead::Nonce {
 ///
 /// Governs version-specific behavior in the TLS layer
 #[non_exhaustive]
-#[derive(Clone, Copy)]
+#[derive(Clone, Copy, Debug)]
 pub enum Version {
     /// Draft versions 29, 30, 31 and 32
     V1Draft,
@@ -483,12 +832,12 @@ pub enum Version {
 impl Version {
     fn initial_salt(self) -> &'static [u8; 20] {
         match self {
-            Version::V1Draft => &[
+            Self::V1Draft => &[
                 // https://datatracker.ietf.org/doc/html/draft-ietf-quic-tls-32#section-5.2
                 0xaf, 0xbf, 0xec, 0x28, 0x99, 0x93, 0xd2, 0x4c, 0x9e, 0x97, 0x86, 0xf1, 0x9c, 0x61,
                 0x11, 0xe0, 0x43, 0x90, 0xa8, 0x99,
             ],
-            Version::V1 => &[
+            Self::V1 => &[
                 // https://www.rfc-editor.org/rfc/rfc9001.html#name-initial-secrets
                 0x38, 0x76, 0x2c, 0xf7, 0xf5, 0x59, 0x34, 0xb3, 0x4d, 0x17, 0x9a, 0xe6, 0xa4, 0xc8,
                 0x0c, 0xad, 0xcc, 0xbb, 0x7f, 0x0a,
@@ -589,7 +938,7 @@ mod test {
                 ],
             ),
             suite: TLS13_AES_128_GCM_SHA256_INTERNAL,
-            is_client: true,
+            side: Side::Client,
         };
         secrets.update();
 
