@@ -1,14 +1,15 @@
-use crate::conn::{CommonState, ConnectionRandoms, State};
-use crate::error::Error;
+use crate::common_state::{CommonState, State};
+use crate::conn::ConnectionRandoms;
+#[cfg(feature = "tls12")]
+use crate::enums::CipherSuite;
+use crate::enums::{AlertDescription, HandshakeType, ProtocolVersion, SignatureScheme};
+use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::{HandshakeHash, HandshakeHashBuffer};
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace};
+use crate::msgs::enums::{Compression, ExtensionType};
 #[cfg(feature = "tls12")]
-use crate::msgs::enums::CipherSuite;
-use crate::msgs::enums::{AlertDescription, Compression, ExtensionType};
-use crate::msgs::enums::{HandshakeType, ProtocolVersion, SignatureScheme};
-#[cfg(feature = "tls12")]
-use crate::msgs::handshake::SessionID;
+use crate::msgs::handshake::SessionId;
 use crate::msgs::handshake::{ClientHelloPayload, Random, ServerExtension};
 use crate::msgs::handshake::{ConvertProtocolNameList, ConvertServerNameList, HandshakePayload};
 use crate::msgs::message::{Message, MessagePayload};
@@ -27,21 +28,21 @@ use std::sync::Arc;
 
 pub(super) type NextState = Box<dyn State<ServerConnectionData>>;
 pub(super) type NextStateOrError = Result<NextState, Error>;
-pub(super) type ServerContext<'a> = crate::conn::Context<'a, ServerConnectionData>;
+pub(super) type ServerContext<'a> = crate::common_state::Context<'a, ServerConnectionData>;
 
-pub(super) fn incompatible(common: &mut CommonState, why: &str) -> Error {
+pub(super) fn incompatible(common: &mut CommonState, why: PeerIncompatible) -> Error {
     common.send_fatal_alert(AlertDescription::HandshakeFailure);
-    Error::PeerIncompatibleError(why.to_string())
+    Error::PeerIncompatible(why)
 }
 
-fn bad_version(common: &mut CommonState, why: &str) -> Error {
+fn bad_version(common: &mut CommonState, why: PeerIncompatible) -> Error {
     common.send_fatal_alert(AlertDescription::ProtocolVersion);
-    Error::PeerIncompatibleError(why.to_string())
+    Error::PeerIncompatible(why)
 }
 
-pub(super) fn decode_error(common: &mut CommonState, why: &str) -> Error {
+pub(super) fn decode_error(common: &mut CommonState, why: PeerMisbehaved) -> Error {
     common.send_fatal_alert(AlertDescription::DecodeError);
-    Error::PeerMisbehavedError(why.to_string())
+    Error::PeerMisbehaved(why)
 }
 
 pub(super) fn can_resume(
@@ -95,9 +96,7 @@ impl ExtensionProcessing {
                 .iter()
                 .any(|protocol| protocol.is_empty())
             {
-                return Err(Error::PeerMisbehavedError(
-                    "client offered empty ALPN protocol".to_string(),
-                ));
+                return Err(PeerMisbehaved::OfferedEmptyApplicationProtocol.into());
             }
 
             cx.common.alpn_protocol = our_protocols
@@ -138,7 +137,7 @@ impl ExtensionProcessing {
                     None => {
                         return Err(cx
                             .common
-                            .missing_extension("QUIC transport parameters not found"));
+                            .missing_extension(PeerMisbehaved::MissingQuicTransportParameters));
                     }
                 }
             }
@@ -239,11 +238,11 @@ pub(super) struct ExpectClientHello {
     pub(super) extra_exts: Vec<ServerExtension>,
     pub(super) transcript: HandshakeHashOrBuffer,
     #[cfg(feature = "tls12")]
-    pub(super) session_id: SessionID,
+    pub(super) session_id: SessionId,
     #[cfg(feature = "tls12")]
     pub(super) using_ems: bool,
     pub(super) done_retry: bool,
-    pub(super) send_ticket: bool,
+    pub(super) send_tickets: usize,
 }
 
 impl ExpectClientHello {
@@ -259,18 +258,18 @@ impl ExpectClientHello {
             extra_exts,
             transcript: HandshakeHashOrBuffer::Buffer(transcript_buffer),
             #[cfg(feature = "tls12")]
-            session_id: SessionID::empty(),
+            session_id: SessionId::empty(),
             #[cfg(feature = "tls12")]
             using_ems: false,
             done_retry: false,
-            send_ticket: false,
+            send_tickets: 0,
         }
     }
 
     /// Continues handling of a `ClientHello` message once config and certificate are available.
     pub(super) fn with_certified_key(
         self,
-        sig_schemes: Vec<SignatureScheme>,
+        mut sig_schemes: Vec<SignatureScheme>,
         client_hello: &ClientHelloPayload,
         m: &Message,
         cx: &mut ServerContext<'_>,
@@ -288,26 +287,29 @@ impl ExpectClientHello {
             if versions.contains(&ProtocolVersion::TLSv1_3) && tls13_enabled {
                 ProtocolVersion::TLSv1_3
             } else if !versions.contains(&ProtocolVersion::TLSv1_2) || !tls12_enabled {
-                return Err(bad_version(cx.common, "TLS1.2 not offered/enabled"));
+                return Err(bad_version(
+                    cx.common,
+                    PeerIncompatible::Tls12NotOfferedOrEnabled,
+                ));
             } else if cx.common.is_quic() {
                 return Err(bad_version(
                     cx.common,
-                    "Expecting QUIC connection, but client does not support TLSv1_3",
+                    PeerIncompatible::Tls13RequiredForQuic,
                 ));
             } else {
                 ProtocolVersion::TLSv1_2
             }
         } else if client_hello.client_version.get_u16() < ProtocolVersion::TLSv1_2.get_u16() {
-            return Err(bad_version(cx.common, "Client does not support TLSv1_2"));
+            return Err(bad_version(cx.common, PeerIncompatible::Tls12NotOffered));
         } else if !tls12_enabled && tls13_enabled {
             return Err(bad_version(
                 cx.common,
-                "Server requires TLS1.3, but client omitted versions ext",
+                PeerIncompatible::SupportedVersionsExtensionRequired,
             ));
         } else if cx.common.is_quic() {
             return Err(bad_version(
                 cx.common,
-                "Expecting QUIC connection, but client does not support TLSv1_3",
+                PeerIncompatible::Tls13RequiredForQuic,
             ));
         } else {
             ProtocolVersion::TLSv1_2
@@ -315,12 +317,33 @@ impl ExpectClientHello {
 
         cx.common.negotiated_version = Some(version);
 
+        // We communicate to the upper layer what kind of key they should choose
+        // via the sigschemes value.  Clients tend to treat this extension
+        // orthogonally to offered ciphersuites (even though, in TLS1.2 it is not).
+        // So: reduce the offered sigschemes to those compatible with the
+        // intersection of ciphersuites.
+        let client_suites = self
+            .config
+            .cipher_suites
+            .iter()
+            .copied()
+            .filter(|scs| {
+                client_hello
+                    .cipher_suites
+                    .contains(&scs.suite())
+            })
+            .collect::<Vec<_>>();
+
+        sig_schemes
+            .retain(|scheme| suites::compatible_sigscheme_for_suites(*scheme, &client_suites));
+
         // Choose a certificate.
         let certkey = {
             let client_hello = ClientHello::new(
                 &cx.data.sni,
                 &sig_schemes,
                 client_hello.get_alpn_extension(),
+                &client_hello.cipher_suites,
             );
 
             let certkey = self
@@ -355,7 +378,7 @@ impl ExpectClientHello {
                 &suitable_suites,
             )
         }
-        .ok_or_else(|| incompatible(cx.common, "no ciphersuites in common"))?;
+        .ok_or_else(|| incompatible(cx.common, PeerIncompatible::NoCipherSuitesInCommon))?;
 
         debug!("decided upon suite {:?}", suite);
         cx.common.suite = Some(suite);
@@ -368,7 +391,7 @@ impl ExpectClientHello {
             _ => {
                 return Err(cx
                     .common
-                    .illegal_param("hash differed on retry"));
+                    .illegal_param(PeerMisbehaved::HandshakeHashVariedAfterRetry));
             }
         };
 
@@ -381,7 +404,7 @@ impl ExpectClientHello {
                 suite,
                 randoms,
                 done_retry: self.done_retry,
-                send_ticket: self.send_ticket,
+                send_tickets: self.send_tickets,
                 extra_exts: self.extra_exts,
             }
             .handle_client_hello(cx, certkey, m, client_hello, sig_schemes),
@@ -393,7 +416,7 @@ impl ExpectClientHello {
                 suite,
                 using_ems: self.using_ems,
                 randoms,
-                send_ticket: self.send_ticket,
+                send_ticket: self.send_tickets > 0,
                 extra_exts: self.extra_exts,
             }
             .handle_client_hello(
@@ -410,13 +433,7 @@ impl ExpectClientHello {
 
 impl State<ServerConnectionData> for ExpectClientHello {
     fn handle(self: Box<Self>, cx: &mut ServerContext<'_>, m: Message) -> NextStateOrError {
-        let (client_hello, sig_schemes) = process_client_hello(
-            &m,
-            self.done_retry,
-            &self.config.cipher_suites,
-            cx.common,
-            cx.data,
-        )?;
+        let (client_hello, sig_schemes) = process_client_hello(&m, self.done_retry, cx)?;
         self.with_certified_key(sig_schemes, client_hello, &m, cx)
     }
 }
@@ -431,9 +448,7 @@ impl State<ServerConnectionData> for ExpectClientHello {
 pub(super) fn process_client_hello<'a>(
     m: &'a Message,
     done_retry: bool,
-    supported_cipher_suites: &[SupportedCipherSuite],
-    common: &mut CommonState,
-    data: &mut ServerConnectionData,
+    cx: &mut ServerContext,
 ) -> Result<(&'a ClientHelloPayload, Vec<SignatureScheme>), Error> {
     let client_hello =
         require_handshake_msg!(m, HandshakeType::ClientHello, HandshakePayload::ClientHello)?;
@@ -443,18 +458,20 @@ pub(super) fn process_client_hello<'a>(
         .compression_methods
         .contains(&Compression::Null)
     {
-        common.send_fatal_alert(AlertDescription::IllegalParameter);
-        return Err(Error::PeerIncompatibleError(
-            "client did not offer Null compression".to_string(),
-        ));
+        cx.common
+            .send_fatal_alert(AlertDescription::IllegalParameter);
+        return Err(PeerIncompatible::NullCompressionRequired.into());
     }
 
     if client_hello.has_duplicate_extension() {
-        return Err(decode_error(common, "client sent duplicate extensions"));
+        return Err(decode_error(
+            cx.common,
+            PeerMisbehaved::DuplicateClientHelloExtensions,
+        ));
     }
 
     // No handshake messages should follow this one in this flight.
-    common.check_aligned_handshake()?;
+    cx.common.check_aligned_handshake()?;
 
     // Extract and validate the SNI DNS name, if any, before giving it to
     // the cert resolver. In particular, if it is invalid then we should
@@ -465,15 +482,17 @@ pub(super) fn process_client_hello<'a>(
         Some(sni) => {
             if sni.has_duplicate_names_for_type() {
                 return Err(decode_error(
-                    common,
-                    "ClientHello SNI contains duplicate name types",
+                    cx.common,
+                    PeerMisbehaved::DuplicateServerNameTypes,
                 ));
             }
 
             if let Some(hostname) = sni.get_single_hostname() {
                 Some(hostname.into())
             } else {
-                return Err(common.illegal_param("ClientHello SNI did not contain a hostname"));
+                return Err(cx
+                    .common
+                    .illegal_param(PeerMisbehaved::ServerNameMustContainOneHostName));
             }
         }
         None => None,
@@ -483,36 +502,22 @@ pub(super) fn process_client_hello<'a>(
     if let (Some(sni), false) = (&sni, done_retry) {
         // Save the SNI into the session.
         // The SNI hostname is immutable once set.
-        assert!(data.sni.is_none());
-        data.sni = Some(sni.clone())
-    } else if data.sni != sni {
-        return Err(Error::PeerIncompatibleError(
-            "SNI differed on retry".to_string(),
-        ));
+        assert!(cx.data.sni.is_none());
+        cx.data.sni = Some(sni.clone());
+    } else if cx.data.sni != sni {
+        return Err(PeerMisbehaved::ServerNameDifferedOnRetry.into());
     }
 
-    // We communicate to the upper layer what kind of key they should choose
-    // via the sigschemes value.  Clients tend to treat this extension
-    // orthogonally to offered ciphersuites (even though, in TLS1.2 it is not).
-    // So: reduce the offered sigschemes to those compatible with the
-    // intersection of ciphersuites.
-    let client_suites = supported_cipher_suites
-        .iter()
-        .copied()
-        .filter(|scs| {
-            client_hello
-                .cipher_suites
-                .contains(&scs.suite())
-        })
-        .collect::<Vec<_>>();
-
-    let mut sig_schemes = client_hello
+    let sig_schemes = client_hello
         .get_sigalgs_extension()
-        .ok_or_else(|| incompatible(common, "client didn't describe signature schemes"))?
-        .clone();
-    sig_schemes.retain(|scheme| suites::compatible_sigscheme_for_suites(*scheme, &client_suites));
+        .ok_or_else(|| {
+            incompatible(
+                cx.common,
+                PeerIncompatible::SignatureAlgorithmsExtensionRequired,
+            )
+        })?;
 
-    Ok((client_hello, sig_schemes))
+    Ok((client_hello, sig_schemes.to_owned()))
 }
 
 #[allow(clippy::large_enum_variant)]
