@@ -1,10 +1,11 @@
+use crate::enums::{SignatureAlgorithm, SignatureScheme};
 use crate::error::Error;
 use crate::key;
-use crate::msgs::enums::{SignatureAlgorithm, SignatureScheme};
+use crate::x509::{wrap_in_asn1_len, wrap_in_sequence};
 
+use ring::io::der;
 use ring::signature::{self, EcdsaKeyPair, Ed25519KeyPair, RsaKeyPair};
 
-use std::convert::TryFrom;
 use std::error::Error as StdError;
 use std::fmt;
 use std::sync::Arc;
@@ -105,7 +106,7 @@ impl CertifiedKey {
             // that the certificate is valid for, if the certificate is
             // valid.
             if end_entity_cert
-                .verify_is_valid_for_dns_name(name)
+                .verify_is_valid_for_subject_name(webpki::SubjectNameRef::DnsName(name))
                 .is_err()
             {
                 return Err(Error::General(
@@ -133,6 +134,9 @@ pub fn any_supported_type(der: &key::PrivateKey) -> Result<Arc<dyn SigningKey>,
 }
 
 /// Parse `der` as any ECDSA key type, returning the first which works.
+///
+/// Both SEC1 (PEM section starting with 'BEGIN EC PRIVATE KEY') and PKCS8
+/// (PEM section starting with 'BEGIN PRIVATE KEY') encodings are supported.
 pub fn any_ecdsa_type(der: &key::PrivateKey) -> Result<Arc<dyn SigningKey>, SignError> {
     if let Ok(ecdsa_p256) = EcdsaSigningKey::new(
         der,
@@ -164,7 +168,11 @@ pub fn any_eddsa_type(der: &key::PrivateKey) -> Result<Arc<dyn SigningKey>, Sign
     Err(SignError(()))
 }
 
-/// A `SigningKey` for RSA-PKCS1 or RSA-PSS
+/// A `SigningKey` for RSA-PKCS1 or RSA-PSS.
+///
+/// This is used by the test suite, so it must be `pub`, but it isn't part of
+/// the public, stable, API.
+#[doc(hidden)]
 pub struct RsaSigningKey {
     key: Arc<RsaKeyPair>,
 }
@@ -179,7 +187,7 @@ static ALL_RSA_SCHEMES: &[SignatureScheme] = &[
 ];
 
 impl RsaSigningKey {
-    /// Make a new `RSASigningKey` from a DER encoding, in either
+    /// Make a new `RsaSigningKey` from a DER encoding, in either
     /// PKCS#1 or PKCS#8 format.
     pub fn new(der: &key::PrivateKey) -> Result<Self, SignError> {
         RsaKeyPair::from_der(&der.0)
@@ -202,11 +210,6 @@ impl SigningKey for RsaSigningKey {
     }
 }
 
-#[allow(clippy::upper_case_acronyms)]
-#[doc(hidden)]
-#[deprecated(since = "0.20.0", note = "Use RsaSigningKey")]
-pub type RSASigningKey = RsaSigningKey;
-
 struct RsaSigner {
     key: Arc<RsaKeyPair>,
     scheme: SignatureScheme,
@@ -252,7 +255,7 @@ impl Signer for RsaSigner {
 /// A SigningKey that uses exactly one TLS-level SignatureScheme
 /// and one ring-level signature::SigningAlgorithm.
 ///
-/// Compare this to RSASigningKey, which for a particular key is
+/// Compare this to RsaSigningKey, which for a particular key is
 /// willing to sign with several algorithms.  This is quite poor
 /// cryptography practice, but is necessary because a given RSA key
 /// is expected to work in TLS1.2 (PKCS#1 signatures) and TLS1.3
@@ -266,22 +269,72 @@ struct EcdsaSigningKey {
 }
 
 impl EcdsaSigningKey {
-    /// Make a new `ECDSASigningKey` from a DER encoding in PKCS#8 format,
-    /// expecting a key usable with precisely the given signature scheme.
+    /// Make a new `ECDSASigningKey` from a DER encoding in PKCS#8 or SEC1
+    /// format, expecting a key usable with precisely the given signature
+    /// scheme.
     fn new(
         der: &key::PrivateKey,
         scheme: SignatureScheme,
         sigalg: &'static signature::EcdsaSigningAlgorithm,
     ) -> Result<Self, ()> {
         EcdsaKeyPair::from_pkcs8(sigalg, &der.0)
+            .map_err(|_| ())
+            .or_else(|_| Self::convert_sec1_to_pkcs8(scheme, sigalg, &der.0))
             .map(|kp| Self {
                 key: Arc::new(kp),
                 scheme,
             })
-            .map_err(|_| ())
+    }
+
+    /// Convert a SEC1 encoding to PKCS8, and ask ring to parse it.  This
+    /// can be removed once https://github.com/briansmith/ring/pull/1456
+    /// (or equivalent) is landed.
+    fn convert_sec1_to_pkcs8(
+        scheme: SignatureScheme,
+        sigalg: &'static signature::EcdsaSigningAlgorithm,
+        maybe_sec1_der: &[u8],
+    ) -> Result<EcdsaKeyPair, ()> {
+        let pkcs8_prefix = match scheme {
+            SignatureScheme::ECDSA_NISTP256_SHA256 => &PKCS8_PREFIX_ECDSA_NISTP256,
+            SignatureScheme::ECDSA_NISTP384_SHA384 => &PKCS8_PREFIX_ECDSA_NISTP384,
+            _ => unreachable!(), // all callers are in this file
+        };
+
+        // wrap sec1 encoding in an OCTET STRING
+        let mut sec1_wrap = Vec::with_capacity(maybe_sec1_der.len() + 8);
+        sec1_wrap.extend_from_slice(maybe_sec1_der);
+        wrap_in_asn1_len(&mut sec1_wrap);
+        sec1_wrap.insert(0, der::Tag::OctetString as u8);
+
+        let mut pkcs8 = Vec::with_capacity(pkcs8_prefix.len() + sec1_wrap.len() + 4);
+        pkcs8.extend_from_slice(pkcs8_prefix);
+        pkcs8.extend_from_slice(&sec1_wrap);
+        wrap_in_sequence(&mut pkcs8);
+
+        EcdsaKeyPair::from_pkcs8(sigalg, &pkcs8).map_err(|_| ())
     }
 }
 
+// This is (line-by-line):
+// - INTEGER Version = 0
+// - SEQUENCE (privateKeyAlgorithm)
+//   - id-ecPublicKey OID
+//   - prime256v1 OID
+const PKCS8_PREFIX_ECDSA_NISTP256: &[u8] = b"\x02\x01\x00\
+      \x30\x13\
+      \x06\x07\x2a\x86\x48\xce\x3d\x02\x01\
+      \x06\x08\x2a\x86\x48\xce\x3d\x03\x01\x07";
+
+// This is (line-by-line):
+// - INTEGER Version = 0
+// - SEQUENCE (privateKeyAlgorithm)
+//   - id-ecPublicKey OID
+//   - secp384r1 OID
+const PKCS8_PREFIX_ECDSA_NISTP384: &[u8] = b"\x02\x01\x00\
+     \x30\x10\
+     \x06\x07\x2a\x86\x48\xce\x3d\x02\x01\
+     \x06\x05\x2b\x81\x04\x00\x22";
+
 impl SigningKey for EcdsaSigningKey {
     fn choose_scheme(&self, offered: &[SignatureScheme]) -> Option<Box<dyn Signer>> {
         if offered.contains(&self.scheme) {
@@ -295,7 +348,6 @@ impl SigningKey for EcdsaSigningKey {
     }
 
     fn algorithm(&self) -> SignatureAlgorithm {
-        use crate::msgs::handshake::DecomposedSignatureScheme;
         self.scheme.sign()
     }
 }
@@ -322,7 +374,7 @@ impl Signer for EcdsaSigner {
 /// A SigningKey that uses exactly one TLS-level SignatureScheme
 /// and one ring-level signature::SigningAlgorithm.
 ///
-/// Compare this to RSASigningKey, which for a particular key is
+/// Compare this to RsaSigningKey, which for a particular key is
 /// willing to sign with several algorithms.  This is quite poor
 /// cryptography practice, but is necessary because a given RSA key
 /// is expected to work in TLS1.2 (PKCS#1 signatures) and TLS1.3
@@ -361,7 +413,6 @@ impl SigningKey for Ed25519SigningKey {
     }
 
     fn algorithm(&self) -> SignatureAlgorithm {
-        use crate::msgs::handshake::DecomposedSignatureScheme;
         self.scheme.sign()
     }
 }
@@ -405,3 +456,59 @@ impl fmt::Display for SignError {
 }
 
 impl StdError for SignError {}
+
+#[test]
+fn can_load_ecdsa_nistp256_pkcs8() {
+    let key = key::PrivateKey(include_bytes!("testdata/nistp256key.pkcs8.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_ecdsa_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_ecdsa_nistp256_sec1() {
+    let key = key::PrivateKey(include_bytes!("testdata/nistp256key.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_ecdsa_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_ecdsa_nistp384_pkcs8() {
+    let key = key::PrivateKey(include_bytes!("testdata/nistp384key.pkcs8.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_ecdsa_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_ecdsa_nistp384_sec1() {
+    let key = key::PrivateKey(include_bytes!("testdata/nistp384key.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_ecdsa_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_eddsa_pkcs8() {
+    let key = key::PrivateKey(include_bytes!("testdata/eddsakey.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_ok());
+    assert!(any_ecdsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_rsa2048_pkcs8() {
+    let key = key::PrivateKey(include_bytes!("testdata/rsa2048key.pkcs8.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+    assert!(any_ecdsa_type(&key).is_err());
+}
+
+#[test]
+fn can_load_rsa2048_pkcs1() {
+    let key = key::PrivateKey(include_bytes!("testdata/rsa2048key.pkcs1.der").to_vec());
+    assert!(any_supported_type(&key).is_ok());
+    assert!(any_eddsa_type(&key).is_err());
+    assert!(any_ecdsa_type(&key).is_err());
+}
