@@ -1,15 +1,18 @@
+use std::fmt;
+
 use crate::anchors::{OwnedTrustAnchor, RootCertStore};
 use crate::client::ServerName;
-use crate::error::Error;
+use crate::enums::SignatureScheme;
+use crate::error::{CertificateError, Error, InvalidMessage, PeerMisbehaved};
 use crate::key::Certificate;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
-use crate::msgs::enums::SignatureScheme;
-use crate::msgs::handshake::{DigitallySignedStruct, DistinguishedNames};
+use crate::msgs::base::PayloadU16;
+use crate::msgs::codec::{Codec, Reader};
+use crate::msgs::handshake::DistinguishedName;
 
 use ring::digest::Digest;
 
-use std::convert::TryFrom;
 use std::sync::Arc;
 use std::time::SystemTime;
 
@@ -44,6 +47,7 @@ static SUPPORTED_SIG_ALGS: SignatureAlgorithms = &[
 
 /// Zero-sized marker type representing verification of a signature.
 #[derive(Debug)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct HandshakeSignatureValid(());
 
 impl HandshakeSignatureValid {
@@ -65,6 +69,7 @@ impl FinishedMessageVerified {
 /// Zero-sized marker type representing verification of a server cert chain.
 #[allow(unreachable_pub)]
 #[derive(Debug)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct ServerCertVerified(());
 
 #[allow(unreachable_pub)]
@@ -77,6 +82,7 @@ impl ServerCertVerified {
 
 /// Zero-sized marker type representing verification of a client cert chain.
 #[derive(Debug)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct ClientCertVerified(());
 
 impl ClientCertVerified {
@@ -89,16 +95,23 @@ impl ClientCertVerified {
 /// Something that can verify a server certificate chain, and verify
 /// signatures made by certificates.
 #[allow(unreachable_pub)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub trait ServerCertVerifier: Send + Sync {
     /// Verify the end-entity certificate `end_entity` is valid for the
     /// hostname `dns_name` and chains to at least one trust anchor.
     ///
-    /// `intermediates` contains the intermediate certificates the client sent
-    /// along with the end-entity certificate; it is in the same order that the
-    /// peer sent them and may be empty.
+    /// `intermediates` contains all certificates other than `end_entity` that
+    /// were sent as part of the server's [Certificate] message. It is in the
+    /// same order that the server sent them and may be empty.
+    ///
+    /// Note that none of the certificates have been parsed yet, so it is the responsibility of
+    /// the implementor to handle invalid data. It is recommended that the implementor returns
+    /// [`Error::InvalidCertificate(CertificateError::BadEncoding)`] when these cases are encountered.
     ///
     /// `scts` contains the Signed Certificate Timestamps (SCTs) the server
-    /// sent with the certificate, if any.
+    /// sent with the end-entity certificate, if any.
+    ///
+    /// [Certificate]: https://datatracker.ietf.org/doc/html/rfc8446#section-4.4.2
     fn verify_server_cert(
         &self,
         end_entity: &Certificate,
@@ -115,10 +128,9 @@ pub trait ServerCertVerifier: Send + Sync {
     /// The signature and algorithm are within `dss`.  `cert` contains the
     /// public key to use.
     ///
-    /// `cert` is the same certificate that was previously validated by a
-    /// call to `verify_server_cert`.
+    /// `cert` has already been validated by [`ServerCertVerifier::verify_server_cert`].
     ///
-    /// If and only if the signature is valid, return HandshakeSignatureValid.
+    /// If and only if the signature is valid, return `Ok(HandshakeSignatureValid)`.
     /// Otherwise, return an error -- rustls will send an alert and abort the
     /// connection.
     ///
@@ -146,6 +158,12 @@ pub trait ServerCertVerifier: Send + Sync {
     /// must only validate signatures using public keys on the right curve --
     /// rustls does not enforce this requirement for you.
     ///
+    /// `cert` has already been validated by [`ServerCertVerifier::verify_server_cert`].
+    ///
+    /// If and only if the signature is valid, return `Ok(HandshakeSignatureValid)`.
+    /// Otherwise, return an error -- rustls will send an alert and abort the
+    /// connection.
+    ///
     /// This trait method has a default implementation that uses webpki to verify
     /// the signature.
     fn verify_tls13_signature(
@@ -179,8 +197,15 @@ pub trait ServerCertVerifier: Send + Sync {
     }
 }
 
-/// A type which encapsuates a string that is a syntactically valid DNS name.
-#[derive(Clone, Debug, PartialEq)]
+impl fmt::Debug for dyn ServerCertVerifier {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "dyn ServerCertVerifier")
+    }
+}
+
+/// A type which encapsulates a string that is a syntactically valid DNS name.
+#[derive(Clone, Eq, Hash, PartialEq)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct DnsName(pub(crate) webpki::DnsName);
 
 impl AsRef<str> for DnsName {
@@ -189,8 +214,18 @@ impl AsRef<str> for DnsName {
     }
 }
 
+impl fmt::Debug for DnsName {
+    // Workaround solution for ServerName debug formatting:
+    // Just show the string contents here, as verify::DnsName is only
+    // used in ServerName which has some more verbose debug output
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "{:?}", &self.as_ref())
+    }
+}
+
 /// Something that can verify a client certificate chain
 #[allow(unreachable_pub)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub trait ClientCertVerifier: Send + Sync {
     /// Returns `true` to enable the server to request a client certificate and
     /// `false` to skip requesting a client certificate. Defaults to `true`.
@@ -198,26 +233,38 @@ pub trait ClientCertVerifier: Send + Sync {
         true
     }
 
-    /// Return `Some(true)` to require a client certificate and `Some(false)` to make
-    /// client authentication optional. Return `None` to abort the connection.
+    /// Return `true` to require a client certificate and `false` to make
+    /// client authentication optional.
     /// Defaults to `Some(self.offer_client_auth())`.
-    fn client_auth_mandatory(&self) -> Option<bool> {
-        Some(self.offer_client_auth())
+    fn client_auth_mandatory(&self) -> bool {
+        self.offer_client_auth()
     }
 
-    /// Returns the subject names of the client authentication trust anchors to
+    /// Returns the [Subjects] of the client authentication trust anchors to
     /// share with the client when requesting client authentication.
     ///
-    /// Return `None` to abort the connection. Return an empty `Vec` to continue
-    /// the handshake without passing a list of CA DNs.
-    fn client_auth_root_subjects(&self) -> Option<DistinguishedNames>;
+    /// These must be DER-encoded X.500 distinguished names, per RFC 5280.
+    /// They are sent in the [`certificate_authorities`] extension of a
+    /// [`CertificateRequest`] message.
+    ///
+    /// [Subjects]: https://datatracker.ietf.org/doc/html/rfc5280#section-4.1.2.6
+    /// [`CertificateRequest`]: https://datatracker.ietf.org/doc/html/rfc8446#section-4.3.2
+    /// [`certificate_authorities`]: https://datatracker.ietf.org/doc/html/rfc8446#section-4.2.4
+    ///
+    /// If the return value is empty, no CertificateRequest message will be sent.
+    fn client_auth_root_subjects(&self) -> &[DistinguishedName];
 
-    /// Verify the end-entity certificate `end_entity` is valid for the
-    /// and chains to at least one of the trust anchors in `roots`.
+    /// Verify the end-entity certificate `end_entity` is valid, acceptable,
+    /// and chains to at least one of the trust anchors trusted by
+    /// this verifier.
     ///
     /// `intermediates` contains the intermediate certificates the
     /// client sent along with the end-entity certificate; it is in the same
     /// order that the peer sent them and may be empty.
+    ///
+    /// Note that none of the certificates have been parsed yet, so it is the responsibility of
+    /// the implementor to handle invalid data. It is recommended that the implementor returns
+    /// [`Error::InvalidCertificate(CertificateError::BadEncoding)`] when these cases are encountered.
     fn verify_client_cert(
         &self,
         end_entity: &Certificate,
@@ -225,16 +272,15 @@ pub trait ClientCertVerifier: Send + Sync {
         now: SystemTime,
     ) -> Result<ClientCertVerified, Error>;
 
-    /// Verify a signature allegedly by the given server certificate.
+    /// Verify a signature allegedly by the given client certificate.
     ///
     /// `message` is not hashed, and needs hashing during the verification.
     /// The signature and algorithm are within `dss`.  `cert` contains the
     /// public key to use.
     ///
-    /// `cert` is the same certificate that was previously validated by a
-    /// call to `verify_server_cert`.
+    /// `cert` has already been validated by [`ClientCertVerifier::verify_client_cert`].
     ///
-    /// If and only if the signature is valid, return HandshakeSignatureValid.
+    /// If and only if the signature is valid, return `Ok(HandshakeSignatureValid)`.
     /// Otherwise, return an error -- rustls will send an alert and abort the
     /// connection.
     ///
@@ -253,12 +299,13 @@ pub trait ClientCertVerifier: Send + Sync {
         verify_signed_struct(message, cert, dss)
     }
 
-    /// Verify a signature allegedly by the given server certificate.
+    /// Verify a signature allegedly by the given client certificate.
     ///
     /// This method is only called for TLS1.3 handshakes.
     ///
-    /// This method is very similar to `verify_tls12_signature`: but note the
-    /// tighter ECDSA SignatureScheme semantics -- e.g. `SignatureScheme::ECDSA_NISTP256_SHA256`
+    /// This method is very similar to `verify_tls12_signature`, but note the
+    /// tighter ECDSA SignatureScheme semantics in TLS 1.3. For example,
+    /// `SignatureScheme::ECDSA_NISTP256_SHA256`
     /// must only validate signatures using public keys on the right curve --
     /// rustls does not enforce this requirement for you.
     ///
@@ -285,6 +332,12 @@ pub trait ClientCertVerifier: Send + Sync {
     }
 }
 
+impl fmt::Debug for dyn ClientCertVerifier {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "dyn ClientCertVerifier")
+    }
+}
+
 impl ServerCertVerifier for WebPkiVerifier {
     /// Will verify the certificate is valid in the following ways:
     /// - Signed by a  trusted `RootCertStore` CA
@@ -302,8 +355,6 @@ impl ServerCertVerifier for WebPkiVerifier {
         let (cert, chain, trustroots) = prepare(end_entity, intermediates, &self.roots)?;
         let webpki_now = webpki::Time::try_from(now).map_err(|_| Error::FailedToGetCurrentTime)?;
 
-        let ServerName::DnsName(dns_name) = server_name;
-
         let cert = cert
             .verify_is_valid_tls_server_cert(
                 SUPPORTED_SIG_ALGS,
@@ -322,14 +373,28 @@ impl ServerCertVerifier for WebPkiVerifier {
             trace!("Unvalidated OCSP response: {:?}", ocsp_response.to_vec());
         }
 
-        cert.verify_is_valid_for_dns_name(dns_name.0.as_ref())
-            .map_err(pki_error)
-            .map(|_| ServerCertVerified::assertion())
+        match server_name {
+            ServerName::DnsName(dns_name) => {
+                let name = webpki::SubjectNameRef::DnsName(dns_name.0.as_ref());
+                cert.verify_is_valid_for_subject_name(name)
+                    .map_err(pki_error)
+                    .map(|_| ServerCertVerified::assertion())
+            }
+            ServerName::IpAddress(ip_addr) => {
+                let ip_addr = webpki::IpAddr::from(*ip_addr);
+                cert.verify_is_valid_for_subject_name(webpki::SubjectNameRef::IpAddress(
+                    webpki::IpAddrRef::from(&ip_addr),
+                ))
+                .map_err(pki_error)
+                .map(|_| ServerCertVerified::assertion())
+            }
+        }
     }
 }
 
 /// Default `ServerCertVerifier`, see the trait impl for more information.
 #[allow(unreachable_pub)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct WebPkiVerifier {
     roots: RootCertStore,
     ct_policy: Option<CertificateTransparencyPolicy>,
@@ -373,6 +438,7 @@ impl WebPkiVerifier {
 /// certificates will no longer be validated, and a warning message will be logged. The deadline
 /// may vary depending on how often you deploy builds with updated dependencies.
 #[allow(unreachable_pub)]
+#[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
 pub struct CertificateTransparencyPolicy {
     logs: &'static [&'static sct::Log<'static>],
     validation_deadline: SystemTime,
@@ -473,14 +539,30 @@ fn prepare<'a, 'b>(
 /// certificate, without any name checking.
 pub struct AllowAnyAuthenticatedClient {
     roots: RootCertStore,
+    subjects: Vec<DistinguishedName>,
 }
 
 impl AllowAnyAuthenticatedClient {
     /// Construct a new `AllowAnyAuthenticatedClient`.
     ///
     /// `roots` is the list of trust anchors to use for certificate validation.
-    pub fn new(roots: RootCertStore) -> Arc<dyn ClientCertVerifier> {
-        Arc::new(Self { roots })
+    pub fn new(roots: RootCertStore) -> Self {
+        Self {
+            subjects: roots
+                .roots
+                .iter()
+                .map(|r| r.subject().clone())
+                .collect(),
+            roots,
+        }
+    }
+
+    /// Wrap this verifier in an [`Arc`] and coerce it to `dyn ClientCertVerifier`
+    #[inline(always)]
+    pub fn boxed(self) -> Arc<dyn ClientCertVerifier> {
+        // This function is needed because `ClientCertVerifier` is only reachable if the
+        // `dangerous_configuration` feature is enabled, which makes coercing hard to outside users
+        Arc::new(self)
     }
 }
 
@@ -489,8 +571,9 @@ impl ClientCertVerifier for AllowAnyAuthenticatedClient {
         true
     }
 
-    fn client_auth_root_subjects(&self) -> Option<DistinguishedNames> {
-        Some(self.roots.subjects())
+    #[allow(deprecated)]
+    fn client_auth_root_subjects(&self) -> &[DistinguishedName] {
+        &self.subjects
     }
 
     fn verify_client_cert(
@@ -526,10 +609,18 @@ impl AllowAnyAnonymousOrAuthenticatedClient {
     /// Construct a new `AllowAnyAnonymousOrAuthenticatedClient`.
     ///
     /// `roots` is the list of trust anchors to use for certificate validation.
-    pub fn new(roots: RootCertStore) -> Arc<dyn ClientCertVerifier> {
-        Arc::new(Self {
-            inner: AllowAnyAuthenticatedClient { roots },
-        })
+    pub fn new(roots: RootCertStore) -> Self {
+        Self {
+            inner: AllowAnyAuthenticatedClient::new(roots),
+        }
+    }
+
+    /// Wrap this verifier in an [`Arc`] and coerce it to `dyn ClientCertVerifier`
+    #[inline(always)]
+    pub fn boxed(self) -> Arc<dyn ClientCertVerifier> {
+        // This function is needed because `ClientCertVerifier` is only reachable if the
+        // `dangerous_configuration` feature is enabled, which makes coercing hard to outside users
+        Arc::new(self)
     }
 }
 
@@ -538,11 +629,11 @@ impl ClientCertVerifier for AllowAnyAnonymousOrAuthenticatedClient {
         self.inner.offer_client_auth()
     }
 
-    fn client_auth_mandatory(&self) -> Option<bool> {
-        Some(false)
+    fn client_auth_mandatory(&self) -> bool {
+        false
     }
 
-    fn client_auth_root_subjects(&self) -> Option<DistinguishedNames> {
+    fn client_auth_root_subjects(&self) -> &[DistinguishedName] {
         self.inner.client_auth_root_subjects()
     }
 
@@ -560,12 +651,16 @@ impl ClientCertVerifier for AllowAnyAnonymousOrAuthenticatedClient {
 fn pki_error(error: webpki::Error) -> Error {
     use webpki::Error::*;
     match error {
-        BadDer | BadDerTime => Error::InvalidCertificateEncoding,
-        InvalidSignatureForPublicKey => Error::InvalidCertificateSignature,
-        UnsupportedSignatureAlgorithm | UnsupportedSignatureAlgorithmForPublicKey => {
-            Error::InvalidCertificateSignatureType
-        }
-        e => Error::InvalidCertificateData(format!("invalid peer certificate: {}", e)),
+        BadDer | BadDerTime => CertificateError::BadEncoding.into(),
+        CertNotValidYet => CertificateError::NotValidYet.into(),
+        CertExpired | InvalidCertValidity => CertificateError::Expired.into(),
+        UnknownIssuer => CertificateError::UnknownIssuer.into(),
+        CertNotValidForName => CertificateError::NotValidForName.into(),
+
+        InvalidSignatureForPublicKey
+        | UnsupportedSignatureAlgorithm
+        | UnsupportedSignatureAlgorithmForPublicKey => CertificateError::BadSignature.into(),
+        _ => CertificateError::Other(Arc::new(error)).into(),
     }
 }
 
@@ -573,9 +668,13 @@ fn pki_error(error: webpki::Error) -> Error {
 pub struct NoClientAuth;
 
 impl NoClientAuth {
-    /// Constructs a `NoClientAuth` and wraps it in an `Arc`.
-    pub fn new() -> Arc<dyn ClientCertVerifier> {
-        Arc::new(NoClientAuth)
+    /// Construct a [`NoClientAuth`], wrap it in an [`Arc`] and coerce it to
+    /// `dyn ClientCertVerifier`.
+    #[inline(always)]
+    pub fn boxed() -> Arc<dyn ClientCertVerifier> {
+        // This function is needed because `ClientCertVerifier` is only reachable if the
+        // `dangerous_configuration` feature is enabled, which makes coercing hard to outside users
+        Arc::new(Self)
     }
 }
 
@@ -584,7 +683,7 @@ impl ClientCertVerifier for NoClientAuth {
         false
     }
 
-    fn client_auth_root_subjects(&self) -> Option<DistinguishedNames> {
+    fn client_auth_root_subjects(&self) -> &[DistinguishedName] {
         unimplemented!();
     }
 
@@ -598,6 +697,43 @@ impl ClientCertVerifier for NoClientAuth {
     }
 }
 
+/// This type combines a [`SignatureScheme`] and a signature payload produced with that scheme.
+#[derive(Debug, Clone)]
+pub struct DigitallySignedStruct {
+    /// The [`SignatureScheme`] used to produce the signature.
+    pub scheme: SignatureScheme,
+    sig: PayloadU16,
+}
+
+impl DigitallySignedStruct {
+    pub(crate) fn new(scheme: SignatureScheme, sig: Vec<u8>) -> Self {
+        Self {
+            scheme,
+            sig: PayloadU16::new(sig),
+        }
+    }
+
+    /// Get the signature.
+    pub fn signature(&self) -> &[u8] {
+        &self.sig.0
+    }
+}
+
+impl Codec for DigitallySignedStruct {
+    #![allow(deprecated)]
+    fn encode(&self, bytes: &mut Vec<u8>) {
+        self.scheme.encode(bytes);
+        self.sig.encode(bytes);
+    }
+
+    fn read(r: &mut Reader) -> Result<Self, InvalidMessage> {
+        let scheme = SignatureScheme::read(r)?;
+        let sig = PayloadU16::read(r)?;
+
+        Ok(Self { scheme, sig })
+    }
+}
+
 static ECDSA_SHA256: SignatureAlgorithms =
     &[&webpki::ECDSA_P256_SHA256, &webpki::ECDSA_P384_SHA256];
 
@@ -629,10 +765,7 @@ fn convert_scheme(scheme: SignatureScheme) -> Result<SignatureAlgorithms, Error>
         SignatureScheme::RSA_PSS_SHA384 => Ok(RSA_PSS_SHA384),
         SignatureScheme::RSA_PSS_SHA512 => Ok(RSA_PSS_SHA512),
 
-        _ => {
-            let error_msg = format!("received unadvertised sig scheme {:?}", scheme);
-            Err(Error::PeerMisbehavedError(error_msg))
-        }
+        _ => Err(PeerMisbehaved::SignedHandshakeWithUnadvertisedSigScheme.into()),
     }
 }
 
@@ -662,7 +795,7 @@ fn verify_signed_struct(
     let possible_algs = convert_scheme(dss.scheme)?;
     let cert = webpki::EndEntityCert::try_from(cert.0.as_ref()).map_err(pki_error)?;
 
-    verify_sig_using_any_alg(&cert, possible_algs, message, &dss.sig.0)
+    verify_sig_using_any_alg(&cert, possible_algs, message, dss.signature())
         .map_err(pki_error)
         .map(|_| HandshakeSignatureValid::assertion())
 }
@@ -670,7 +803,7 @@ fn verify_signed_struct(
 fn convert_alg_tls13(
     scheme: SignatureScheme,
 ) -> Result<&'static webpki::SignatureAlgorithm, Error> {
-    use crate::msgs::enums::SignatureScheme::*;
+    use crate::enums::SignatureScheme::*;
 
     match scheme {
         ECDSA_NISTP256_SHA256 => Ok(&webpki::ECDSA_P256_SHA256),
@@ -679,10 +812,7 @@ fn convert_alg_tls13(
         RSA_PSS_SHA256 => Ok(&webpki::RSA_PSS_2048_8192_SHA256_LEGACY_KEY),
         RSA_PSS_SHA384 => Ok(&webpki::RSA_PSS_2048_8192_SHA384_LEGACY_KEY),
         RSA_PSS_SHA512 => Ok(&webpki::RSA_PSS_2048_8192_SHA512_LEGACY_KEY),
-        _ => {
-            let error_msg = format!("received unsupported sig scheme {:?}", scheme);
-            Err(Error::PeerMisbehavedError(error_msg))
-        }
+        _ => Err(PeerMisbehaved::SignedHandshakeWithUnadvertisedSigScheme.into()),
     }
 }
 
@@ -716,7 +846,7 @@ fn verify_tls13(
 
     let cert = webpki::EndEntityCert::try_from(cert.0.as_ref()).map_err(pki_error)?;
 
-    cert.verify_signature(alg, msg, &dss.sig.0)
+    cert.verify_signature(alg, msg, dss.signature())
         .map_err(pki_error)
         .map(|_| HandshakeSignatureValid::assertion())
 }
